# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import io
import random
import time
from typing import Optional, List

from jinja2 import Template

from kubemarine import system, packages
from kubemarine.core import utils, static
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup, NodeConfig, RunnersGroupResult


def autodetect_interface(cluster: KubernetesCluster, name: str) -> Optional[str]:
    for node in cluster.inventory['nodes']:
        if node['name'] == name:
            address = cluster.get_access_address_from_node(node)
            interface: str = cluster.context['nodes'].get(address, {})['active_interface']
            if interface:
                return interface
    if cluster.context['initial_procedure'] == 'remove_node':
        for node_to_remove in cluster.procedure_inventory['nodes']:
            if node_to_remove['name'] == name:
                return None
    raise Exception('Failed to autodetect active interface for %s' % name)


def enrich_inventory_apply_defaults(inventory: dict, cluster: KubernetesCluster) -> dict:
    # if vrrp_ips is empty, then nothing to do
    if not inventory['vrrp_ips']:
        return inventory

    default_names = get_default_node_names(inventory)

    cluster.log.verbose("Detected default keepalived hosts: %s" % default_names)
    if not default_names:
        cluster.log.verbose("WARNING: Default keepalived hosts are empty: something can go wrong!")

    # iterate over each vrrp_ips item and check if any hosts defined to be used in it
    for i, item in enumerate(inventory['vrrp_ips']):

        if isinstance(item, str):
            inventory['vrrp_ips'][i] = item = {
                'ip': item
            }

        # is router_id defined?
        if item.get('router_id') is None:
            # is there ipv6?
            if ':' in item['ip']:
                item['router_id'] = item['ip'].split(':').pop()
                if item['router_id'] == '':
                    item['router_id'] = '0'
                # in adress with long last octet e.g. "765d" it is necessary to use only last "5d" and convert it from hex to int
                item['router_id'] = str(int(item['router_id'][-2:], 16))
            else:
                item['router_id'] = item['ip'].split('.').pop()

        # is id defined?
        if item.get('id') is None:
            # label max size is 15, then 15 - 5 (size from string "vip_") = 10 symbols we can use
            source_string = item.get('interface', 'auto') + item['ip']
            label_size = cluster.globals['keepalived']['defaults']['label_size']
            item['id'] = hashlib.md5(source_string.encode('utf-8')).hexdigest()[:label_size]

        # is password defined?
        if item.get('password') is None:
            password_size = cluster.globals['keepalived']['defaults']['password_size']
            item['password'] = ("%032x" % random.getrandbits(128))[:password_size]

        # if nothing defined then use default names
        if item.get('hosts') is None:
            # is there default names found?
            if not default_names:
                raise Exception('Section #%s in vrrp_ips has no hosts, but default names can\'t be found.' % i)
            # ok, default names found, and can be used
            inventory['vrrp_ips'][i]['hosts'] = default_names

        for j, record in enumerate(item['hosts']):
            if isinstance(record, str):
                item['hosts'][j] = {
                    'name': record
                }
            if not item['hosts'][j].get('priority'):
                item['hosts'][j]['priority'] = cluster.globals['keepalived']['defaults']['priority']['max_value'] - \
                                               (j + cluster.globals['keepalived']['defaults']['priority']['step'])
            if not item['hosts'][j].get('interface') and item.get('interface'):
                item['hosts'][j]['interface'] = item['interface']
            if item['hosts'][j].get('interface', 'auto') == 'auto':
                item['hosts'][j]['interface'] = autodetect_interface(cluster, item['hosts'][j]['name'])

    return inventory


def get_default_node_names(inventory: dict) -> List[str]:
    default_names = []

    # well, vrrp_ips is not empty, let's find balancers defined in config-file
    for i, node in enumerate(inventory['nodes']):
        if 'balancer' in node['roles']:
            default_names.append(node['name'])

    # just in case, we remove duplicates
    return list(set(default_names))


def enrich_inventory_calculate_nodegroup(inventory: dict, cluster: KubernetesCluster) -> dict:
    # if vrrp_ips is empty, then nothing to do
    if not inventory['vrrp_ips']:
        return inventory

    # Calculate group, where keepalived should be installed:
    names = []

    for i, item in enumerate(cluster.inventory['vrrp_ips']):
        for record in item['hosts']:
            names.append(record['name'])

    # it is important to remove duplicates
    names = list(set(names))

    # create new group where keepalived will be installed
    cluster.nodes['keepalived'] = cluster.nodes['all'].new_group(apply_filter={
        'name': names
    })

    # create new role
    cluster.roles.append('keepalived')

    # fill in ips
    cluster.ips['keepalived'] = cluster.nodes['keepalived'].get_hosts()

    return inventory


def install(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    # todo why check and try to install all keepalives but finally filter out only new nodes?
    group = group.get_new_nodes_or_self()
    # todo consider probably different associations for nodes with different OS families
    any_host = group.get_first_member().get_host()
    package_associations = cluster.get_associations_for_node(any_host, 'keepalived')

    keepalived_version = group.sudo("%s -v" % package_associations['executable_name'], warn=True)
    keepalived_installed = True

    for connection, result in keepalived_version.items():
        if result.exited != 0:
            keepalived_installed = False

    if keepalived_installed:
        log.debug("Keepalived already installed, nothing to install")
        installation_result = keepalived_version
    else:
        installation_result = packages.install(group, include=package_associations['package_name'])

    service_name = package_associations['service_name']
    patch_path = "./resources/drop_ins/keepalived.conf"
    group.call(system.patch_systemd_service, service_name=service_name, patch_source=patch_path)
    group.call(install_haproxy_check_script)
    enable(group)

    return installation_result


def install_haproxy_check_script(group: NodeGroup) -> None:
    script = utils.read_internal("./resources/scripts/check_haproxy.sh")
    group.put(io.StringIO(script), "/usr/local/bin/check_haproxy.sh", sudo=True)
    group.sudo("chmod +x /usr/local/bin/check_haproxy.sh")


def uninstall(group: NodeGroup) -> RunnersGroupResult:
    return packages.remove(group, include='keepalived')


def restart(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    cluster.log.debug("Restarting keepalived in all group...")
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            service_name = cluster.get_package_association_for_node(
                node.get_host(), 'keepalived', 'service_name')
            system.restart_service(node, name=service_name)

    cluster.log.debug("Sleep while keepalived comes-up...")
    time.sleep(static.GLOBALS['keepalived']['restart_wait'])


def enable(group: NodeGroup) -> None:
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            service_name = exe.cluster.get_package_association_for_node(
                node.get_host(), 'keepalived', 'service_name')
            system.enable_service(node, name=service_name, now=True)


def disable(group: NodeGroup) -> None:
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            service_name = exe.cluster.get_package_association_for_node(
                node.get_host(), 'keepalived', 'service_name')
            system.disable_service(node, name=service_name)


def generate_config(inventory: dict, node: NodeConfig) -> str:
    config = ''

    for i, item in enumerate(inventory['vrrp_ips']):

        if i > 0:
            # this is required for double newline in config, but avoid double newline in the end of file
            config += "\n"

        ips = {
            'source': node['internal_address'],
            'peers': []
        }

        priority = 100
        interface = 'eth0'
        # todo Probably skip the VRRP if it not defined for this node?
        #  Currently behaviour does not correspond to documentation.
        for record in item['hosts']:
            if record['name'] == node['name']:
                priority = record['priority']
                interface = record['interface']

        for i_node in inventory['nodes']:
            for record in item['hosts']:
                if i_node['name'] == record['name'] and i_node['internal_address'] != ips['source']:
                    ips['peers'].append(i_node['internal_address'])

        config_source = utils.read_internal('templates/keepalived.conf.j2')
        config += Template(config_source).render(inventory=inventory, item=item, node=node,
                                                 interface=interface,
                                                 priority=priority, **ips) + "\n"

    return config


def configure(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            node_name = node.get_node_name()
            log.debug("Configuring keepalived on '%s'..." % node_name)

            package_associations = cluster.get_associations_for_node(node.get_host(), 'keepalived')
            configs_directory = '/'.join(package_associations['config_location'].split('/')[:-1])

            exe.group.sudo('mkdir -p %s' % configs_directory)

            config = generate_config(cluster.inventory, node.get_config())
            utils.dump_file(cluster, config, 'keepalived_%s.conf' % node_name)

            node.put(io.StringIO(config), package_associations['config_location'], sudo=True)

    log.debug(group.sudo('ls -la %s' % package_associations['config_location']))

    restart(group)

    return group.sudo('systemctl status %s' % package_associations['service_name'], warn=True)



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent
from typing import List, Optional

import yaml

from kubemarine.core import log
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest


def enrich_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if not inventory["plugins"]["local-path-provisioner"]["install"]:
        return inventory

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_plugin = cluster.raw_inventory.get("plugins", {}).get("local-path-provisioner", {})
    if "resources" in raw_plugin:
        inventory["plugins"]["local-path-provisioner"]["resources"] = raw_plugin["resources"]

    return inventory


class LocalPathProvisionerManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None) -> None:
        super().__init__(logger, inventory, 'local-path-provisioner', original_yaml_path, destination_name)

    def get_known_objects(self) -> List[str]:
        return [
            "Namespace_local-path-storage",
            "ServiceAccount_local-path-provisioner-service-account",
            "ClusterRole_local-path-provisioner-role",
            "ClusterRoleBinding_local-path-provisioner-bind",
            "Deployment_local-path-provisioner",
            "StorageClass_local-path",
            "ConfigMap_local-path-config",
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.enrich_namespace_local_path_storage,
            self.add_clusterrolebinding_local_path_provisioner_privileged_psp,
            self.enrich_deployment_local_path_provisioner,
            self.enrich_storageclass_local_path,
            self.enrich_configmap_local_path_config,
        ]

    def enrich_namespace_local_path_storage(self, manifest: Manifest) -> None:
        key = "Namespace_local-path-storage"
        rbac = self.inventory['rbac']
        if rbac['admission'] == 'pss' and rbac['pss']['pod-security'] == 'enabled' \
                and rbac['pss']['defaults']['enforce'] != 'privileged':
            self.assign_default_pss_labels(manifest, key, 'privileged')

    def add_clusterrolebinding_local_path_provisioner_privileged_psp(self, manifest: Manifest) -> None:
        # TODO add only if psp is enabled?
        new_yaml = yaml.safe_load(clusterrolebinding_local_path_provisioner_privileged_psp)
        # Insert new ClusterRoleBinding after all existing resources of this kind
        max_crb_idx = max(i for i, key in enumerate(manifest.all_obj_keys())
                          if key.startswith("ClusterRoleBinding_"))
        self.include(manifest, max_crb_idx + 1, new_yaml)

    def enrich_deployment_local_path_provisioner(self, manifest: Manifest) -> None:
        key = "Deployment_local-path-provisioner"
        self.enrich_image_for_container(manifest, key,
            container_name='local-path-provisioner', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='local-path-provisioner')
        self.enrich_tolerations(manifest, key)

    def enrich_storageclass_local_path(self, manifest: Manifest) -> None:
        key = "StorageClass_local-path"
        source_yaml = manifest.get_obj(key, patch=True)
        metadata = source_yaml['metadata']

        is_default = str(self.inventory['plugins']['local-path-provisioner']['storage-class']['is-default'])
        metadata.setdefault('annotations', {})['storageclass.kubernetes.io/is-default-class'] = is_default
        self.log.verbose(f"The {key} has been patched in 'metadata.annotations' "
                         f"with 'storageclass.kubernetes.io/is-default-class: {is_default}'")

        name = self.inventory['plugins']['local-path-provisioner']['storage-class']['name']
        metadata['name'] = self.inventory['plugins']['local-path-provisioner']['storage-class']['name']
        self.log.verbose(f"The {key} has been patched in 'metadata.name' with {name!r}")

    def enrich_configmap_local_path_config(self, manifest: Manifest) -> None:
        key = "ConfigMap_local-path-config"
        source_yaml = manifest.get_obj(key, patch=True)
        data = source_yaml['data']

        config_json = data['config.json']
        config_json = config_json.replace('/opt/local-path-provisioner',
                                          self.inventory['plugins']['local-path-provisioner']['volume-dir'])
        data['config.json'] = config_json
        config_json_oneline = config_json.replace('\n', ' ')
        self.log.verbose(f"The {key} has been patched in 'data.config.json' with {config_json_oneline!r}")

        helperpod_yaml_str = data['helperPod.yaml']
        helperpod_yaml = yaml.safe_load(helperpod_yaml_str)
        busybox_source_image = helperpod_yaml['spec']['containers'][0]['image']
        helperpod_yaml_str = helperpod_yaml_str.replace(
            busybox_source_image, self.get_target_image(image_key='helper-pod-image'))
        data['helperPod.yaml'] = helperpod_yaml_str
        helperpod_yaml_str_oneline = helperpod_yaml_str.replace('\n', ' ')
        self.log.verbose(f"The {key} has been patched in 'data.helperPod.yaml' with {helperpod_yaml_str_oneline!r}")


clusterrolebinding_local_path_provisioner_privileged_psp = dedent("""\
    apiVersion: rbac.authorization.k8s.io/v1
    kind: ClusterRoleBinding
    metadata:
      name: local-path-provisioner-privileged-psp
    roleRef:
      kind: ClusterRole
      name: oob-privileged-psp-cr
      apiGroup: rbac.authorization.k8s.io
    subjects:
    - kind: ServiceAccount
      name: local-path-provisioner-service-account
      namespace: local-path-storage
""")



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import StringIO

import toml
import yaml
import base64

from distutils.util import strtobool
from kubemarine import system, packages
from kubemarine.core import utils
from kubemarine.core.group import NodeGroup, RunnersGroupResult, CollectorCallback


def install(group: NodeGroup) -> RunnersGroupResult:
    collector = CollectorCallback(group.cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            os_specific_associations = exe.cluster.get_associations_for_node(node.get_host(), 'containerd')

            exe.cluster.log.debug("Installing latest containerd on %s node" % node.get_node_name())
            # always install latest available containerd
            packages.install(node, include=os_specific_associations['package_name'], callback=collector)

            # remove previous config.toml to avoid problems in case when previous config was broken
            node.sudo("rm -f %s && sudo systemctl restart %s"
                      % (os_specific_associations['config_location'],
                         os_specific_associations['service_name']),
                      callback=collector)

            system.enable_service(node, os_specific_associations['service_name'],
                                  now=True, callback=collector)
    return collector.result


def convert_toml_dict_to_str(config: dict) -> str:
    config_string = ''
    # double loop is used to make sure that no "simple" `key: value` pairs are accidentally assigned to sections
    for key, value in config.items():
        # first we process all "simple" `key: value` pairs
        if not isinstance(value, dict):
            config_string += f"{toml.dumps({key: value})}"
    for key, value in config.items():
        # next we process all "complex" `key: dict_value` pairs, representing named sections
        if isinstance(value, dict):
            config_string += f"\n[{key}]\n{toml.dumps(value)}"
    return config_string


def configure_ctr_flags(group: NodeGroup) -> None:
    cluster = group.cluster
    log = cluster.log
    containerd_config = cluster.inventory["services"]["cri"]['containerdConfig']
    config_string = convert_toml_dict_to_str(containerd_config)
    config_toml = toml.loads(config_string)

    # Calculate ctr options for image pull
    registry = config_toml.get('plugins', {}).get('io.containerd.grpc.v1.cri', {}).get('registry', {})
    ctr_pull_options_str = ""
    for registry_name in set().union(registry.get('mirrors', {}).keys(), registry.get('configs', {}).keys()):
        options = []
        # Add plain-http flag for http endpoint
        if any("http://" in endpoint for endpoint in
               registry.get('mirrors', {}).get(registry_name, {}).get('endpoint', [])):
            options.append('--plain-http')
        # Add skip-verify flag in case of insecure tls connection
        if registry.get('configs', {}).get(registry_name, {}).get('tls', {}).get('insecure_skip_verify', False):
            options.append('--skip-verify')
        # Add user flag if authorization required
        registry_auth = registry.get('configs', {}).get(registry_name, {}).get('auth', {})
        if registry_auth.get('auth'):
            options.append(f'--user {base64.b64decode(registry_auth["auth"]).decode("utf-8")}')

        elif registry_auth.get('username'):
            options.append(f'--user {registry_auth["username"]}' +
                           f':{registry_auth["password"]}' if registry_auth.get("password") else '')
        ctr_pull_options_str += f'{registry_name}={" ".join(options)}\n'

    # Save ctr pull options
    log.debug("Uploading ctr flags configuration...")
    group.put(StringIO(ctr_pull_options_str), "/etc/ctr/kubemarine_ctr_flags.conf", backup=True, sudo=True, mkdir=True)
    group.sudo("chmod 600 /etc/ctr/kubemarine_ctr_flags.conf")


def configure(group: NodeGroup) -> RunnersGroupResult:
    cluster = group.cluster
    log = cluster.log

    log.debug("Uploading crictl configuration for containerd...")
    crictl_config = yaml.dump({"runtime-endpoint": "unix:///run/containerd/containerd.sock"})
    utils.dump_file(cluster, crictl_config, 'crictl.yaml')
    group.put(StringIO(crictl_config), '/etc/crictl.yaml', backup=True, sudo=True)

    configure_ctr_flags(group)

    config_string = convert_toml_dict_to_str(cluster.inventory["services"]["cri"]['containerdConfig'])
    utils.dump_file(cluster, config_string, 'containerd-config.toml')
    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            os_specific_associations = exe.cluster.get_associations_for_node(node.get_host(), 'containerd')
            log.debug("Uploading containerd configuration to %s node..." % node.get_node_name())
            node.put(StringIO(config_string), os_specific_associations['config_location'],
                     backup=True, sudo=True, mkdir=True)
            log.debug("Restarting Containerd on %s node..." % node.get_node_name())
            node.sudo(
                f"chmod 600 {os_specific_associations['config_location']} && "
                f"sudo systemctl restart {os_specific_associations['service_name']} && "
                f"systemctl status {os_specific_associations['service_name']}", callback=collector)
    return collector.result


def prune(group: NodeGroup) -> RunnersGroupResult:
    return group.sudo('crictl rm -fa; '
                      'sudo crictl rmp -fa; '
                      'sudo crictl rmi -a; '
                      'sudo ctr content ls -q | xargs -r sudo ctr content rm', warn=True)



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent

from kubemarine.core.action import Action
from kubemarine.core.patch import InventoryOnlyPatch
from kubemarine.core.resources import DynamicResources


class TheAction(Action):
    def __init__(self) -> None:
        super().__init__("Update user inventory to keep backward-compatible resources requests/limits for plugins")

    def run(self, res: DynamicResources) -> None:
        inventory = res.formatted_inventory()

        self.recreate_inventory = True
        plugins = inventory.setdefault("plugins", {})

        local_path_provisioner = plugins.setdefault("local-path-provisioner", {})
        local_path_provisioner["resources"] = {}

        nginx_ingress = plugins.setdefault("nginx-ingress-controller", {})
        nginx_ingress.setdefault("controller", {})["resources"] = {"requests": {"cpu": "100m", "memory": "90Mi"}}
        nginx_ingress.setdefault("webhook", {})["resources"] = {}

        kubernetes_dashboard = plugins.setdefault("kubernetes-dashboard", {})
        kubernetes_dashboard.setdefault("dashboard", {})["resources"] = {}
        kubernetes_dashboard.setdefault("metrics-scraper", {})["resources"] = {}

        calico = plugins.setdefault("calico", {})
        calico.setdefault("node", {})["resources"] = {"requests": {"cpu": "250m"}}
        calico.setdefault("typha", {})["resources"] = {}
        calico.setdefault("kube-controllers", {})["resources"] = {}


class PluginsResourcesPatch(InventoryOnlyPatch):
    def __init__(self) -> None:
        super().__init__("plugins_resources")

    @property
    def action(self) -> Action:
        return TheAction()

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            Update user inventory to keep backward-compatible resources requests/limits for plugins
            """.rstrip()
        )


# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import time
from typing import List, Dict, Optional

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup


# the methods requires etcdctl.sh to be installed on all active control-plane nodes during thirdparties task.

def remove_members(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    control_planes = cluster.nodes["control-plane"]
    managing_control_plane = control_planes.get_unchanged_nodes().get_any_member()

    log.verbose(f"etcd will be managed using {managing_control_plane.get_nodes_names()[0]}.")
    output = managing_control_plane.sudo("etcdctl member list").get_simple_out().splitlines()

    etcd_members = {}
    for line in output:
        params = [p.strip() for p in line.split(sep=',')]
        # 6 is expected number of comma-separated parameters of an etcd member
        if len(params) == 6:
            etcd_members[params[2]] = params[0]
        else:
            log.warning("Unexpected line in 'etcdctl member list' output: " + line)

    log.verbose(f"Found etcd members {list(etcd_members.keys())}")
    unexpected_members = etcd_members.keys() - set(control_planes.get_nodes_names())
    if unexpected_members:
        log.warning(f"Found unexpected etcd members {list(unexpected_members)}")

    for node_name in group.get_nodes_names():
        if node_name in etcd_members:
            command = "etcdctl member remove " + etcd_members[node_name]
            log.verbose(f"Removing found etcd member {node_name}...")
            managing_control_plane.sudo(command)
        else:
            log.verbose(f"Skipping {node_name} as it is not among etcd members.")


def wait_for_health(cluster: KubernetesCluster, connection: NodeGroup) -> List[Dict]:
    """
    The method checks etcd endpoints health until all endpoints are healthy or retries are exhausted
    if all member are healthy the method checks the leader.
    """
    log = cluster.log
    init_timeout = cluster.globals['etcd']['health']['init_timeout']
    timeout = cluster.globals['etcd']['health']['timeout']
    retries = cluster.globals['etcd']['health']['retries']

    is_healthy = False
    time.sleep(init_timeout)
    while retries > 0:
        start_time = time.time()
        etcd_health_raw = connection.sudo('etcdctl endpoint health --cluster -w json').get_simple_out()
        end_time = time.time()
        sudo_time = int(end_time - start_time)
        log.verbose(etcd_health_raw)
        etcd_health_list = json.load(io.StringIO(etcd_health_raw.strip()))

        health = 0
        for etcd_health in etcd_health_list:
            if etcd_health.get('health'):
                health += 1

        if health == len(etcd_health_list):
            log.debug('All ETCD members are healthy!')
            is_healthy = True
            break
        else:
            log.debug('Wait for ETCD cluster is not healthy!')
            if sudo_time < timeout:
                time.sleep(timeout - sudo_time)
            retries -= 1

    if is_healthy:
        etcd_status_raw = connection.sudo('etcdctl endpoint status --cluster -w json').get_simple_out()
        log.verbose(etcd_status_raw)
        etcd_status_list: List[dict] = json.load(io.StringIO(etcd_status_raw.lower().strip()))
        elected_leader: Optional[int] = None
        for item in etcd_status_list:
            leader: Optional[int] = item.get('status', {}).get('leader')
            if leader is None:
                raise Exception('ETCD member "%s" do not have leader' % item.get('endpoint'))
            if elected_leader is None:
                elected_leader = leader
            elif elected_leader != leader:
                raise Exception('ETCD leaders are not the same')
        log.debug('Leader "%s" elected' % elected_leader)
    else:
        raise Exception('ETCD cluster is still not healthy!')
      
    log.verbose('ETCD cluster is healthy!')

    return etcd_status_list



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
from abc import ABC, abstractmethod

from pygelf import gelf, GelfTcpHandler, GelfUdpHandler, GelfTlsHandler, GelfHttpHandler  # type: ignore[import]

from copy import deepcopy
from typing import Any, List, Optional, cast, Dict, Union

VERBOSE = 5
gelf.LEVELS.update({VERBOSE: 8})

DEFAULT_FORMAT = '%(asctime)s %(name)s %(levelname)s %(message)s'

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"

COLORS = {
    'RESET': RESET_SEQ,
    'BOLD': BOLD_SEQ,
    'BLACK': COLOR_SEQ % (30 + BLACK),
    'RED': COLOR_SEQ % (30 + RED),
    'GREEN': COLOR_SEQ % (30 + GREEN),
    'YELLOW': COLOR_SEQ % (30 + YELLOW),
    'BLUE': COLOR_SEQ % (30 + BLUE),
    'MAGENTA': COLOR_SEQ % (30 + MAGENTA),
    'CYAN': COLOR_SEQ % (30 + CYAN),
    'WHITE': COLOR_SEQ % (30 + WHITE),
}

COLORS_SCHEME = {
    'WARNING': 'YELLOW',
    'VERBOSE': 'BLUE',
    'INFO': 'GREEN',
    'ERROR': 'RED',
    'CRITICAL': 'RED'
}

LOGGING_LEVELS_BY_NAME = {
    '5': VERBOSE,
    'verbose': VERBOSE,
    '10': logging.DEBUG,
    'debug': logging.DEBUG,
    '20': logging.INFO,
    'info': logging.INFO,
    '30': logging.WARNING,
    'warn': logging.WARNING,
    'warning': logging.WARNING,
    '40': logging.ERROR,
    'error': logging.ERROR,
    '50': logging.CRITICAL,
    'critical': logging.CRITICAL
}

LOGGING_NAMES_BY_LEVEL = {
    VERBOSE: 'verbose',
    logging.DEBUG: 'debug',
    logging.INFO: 'info',
    logging.WARNING: 'warning',
    logging.ERROR: 'error',
    logging.CRITICAL: 'critical'
}


class VerboseLogger(ABC):
    @abstractmethod
    def verbose(self, msg: object, *args: object, **kwargs: Any) -> None:
        pass


class EnhancedLogger(logging.Logger, VerboseLogger):
    def __init__(self, name: str, level: int = logging.NOTSET):
        super().__init__(name, level)
        logging.addLevelName(VERBOSE, 'VERBOSE')

    def verbose(self, msg: object, *args: object, **kwargs: Any) -> None:
        if self.isEnabledFor(VERBOSE):
            self._log(VERBOSE, msg, args, **kwargs)

    def makeRecord(self, name: str, level: int, fn: str, lno: int, msg: object, args,  # type: ignore[no-untyped-def]
                   exc_info, func=None, extra=None,
                   sinfo=None) -> logging.LogRecord:
        record = super().makeRecord(name, level, fn, lno, msg, args, exc_info, func, extra, sinfo)
        caller = record.__dict__.get('real_caller')
        if caller is not None:
            record.__dict__.update(record.__dict__.pop('real_caller'))

        return record


class EnhancedLogRecord(logging.LogRecord):
    def getMessage(self) -> str:
        message = super().getMessage()
        prefix = self.__dict__.get('prefix')
        if prefix is not None:
            message = prefix + message
        return message


logging.setLoggerClass(EnhancedLogger)
logging.setLogRecordFactory(EnhancedLogRecord)


class LogFormatter(logging.Formatter):
    def __init__(self, fmt: str = None, datefmt: str = None,
                 colorize: bool = False, correct_newlines: bool = False):
        super().__init__(fmt, datefmt)
        self.colorize = colorize
        self.correct_newlines = correct_newlines

    def _format(self, record: logging.LogRecord) -> str:
        s = super().format(record)
        if self.colorize and record.levelname in COLORS_SCHEME:
            s = COLORS[COLORS_SCHEME[record.levelname]] + s + COLORS['RESET']
        return s

    def format(self, record: logging.LogRecord) -> str:
        if self.correct_newlines:
            messages = str(record.msg).split('\n')
            if len(messages) == 1:
                return self._format(record)
        else:
            return self._format(record)

        orig_msg = record.msg
        try:
            s = ''
            for message in messages:
                if s != '':
                    s += '\n'
                record.msg = message
                s += self._format(record)
            return s
        finally:
            record.msg = orig_msg


class StdoutHandler(logging.StreamHandler):
    def __init__(self, formatter: LogFormatter):
        super().__init__(sys.stdout)
        self.formatter: LogFormatter = formatter

    def emit(self, record: logging.LogRecord) -> None:
        if 'ignore_stdout' in record.__dict__:
            return
        super().emit(record)
        # TODO: if output stuck, then add here self.flush()
        # More about, see at https://stackoverflow.com/questions/16633911


class FileHandlerWithHeader(logging.FileHandler):
    def __init__(self, formatter: LogFormatter, filename: str, header: str = None,
                 mode: str = 'a', encoding: str = None):
        # Store the header information.
        self.header = header

        # Call the parent __init__
        logging.FileHandler.__init__(self, filename, mode, encoding)

        # Write the header if it was specified
        if header:
            self.stream.write('%s\n' % header)

        self.formatter: LogFormatter = formatter

    def emit(self, record: logging.LogRecord) -> None:
        # Call the parent class emit function.
        logging.FileHandler.emit(self, record)


class LogHandler:

    def __init__(self,
                 target: str,
                 level: str,
                 colorize: bool = False,
                 correct_newlines: bool = False,
                 filemode: str = 'a',
                 format: str = DEFAULT_FORMAT,
                 datefmt: str = None,
                 header: str = None,
                 **kwargs: Union[str, bool, int]):

        self._colorize = colorize
        self._correct_newlines = correct_newlines
        self._format = format
        self._datefmt = datefmt
        self._header = header

        self._formatter = LogFormatter(self._format, self._datefmt,
                                       colorize=self._colorize,
                                       correct_newlines=self._correct_newlines)

        self.handler: logging.Handler
        if target.lower() == 'stdout':
            self._target = 'stdout'
            self.handler = StdoutHandler(self._formatter)
        elif target.lower() == 'graylog':
            self._target = 'graylog'
            if not kwargs.get('host'):
                raise Exception('Graylog host is not defined')
            if not kwargs.get('port'):
                raise Exception(f'Graylog port is not defined for "{kwargs["host"]}"')
            if not kwargs.get('type'):
                raise Exception(f'Graylog type is not defined for "{kwargs["host"]}:{kwargs["port"]}"')
            handler_options: Dict[str, Union[str, bool, int, None]] = {
                'host': kwargs['host'],
                'port': kwargs['port'],
                '_app_name': kwargs.get('appname', 'kubemarine'),
                'debug': kwargs.get('debug', False),
                'version': kwargs.get('version', '1.1')
            }
            if kwargs['type'] == 'tcp':
                self.handler = GelfTcpHandler(**handler_options)
            elif kwargs['type'] == 'udp':
                handler_options['compress'] = kwargs.get('compress', True)
                handler_options['chunk_size'] = kwargs.get('chunk_size', 1300)
                self.handler = GelfUdpHandler(**handler_options)
            elif kwargs['type'] == 'tls':
                handler_options['validate'] = kwargs.get('validate', True)
                handler_options['ca_certs'] = kwargs.get('ca_certs')
                handler_options['certfile'] = kwargs.get('certfile')
                handler_options['keyfile'] = kwargs.get('keyfile')
                self.handler = GelfTlsHandler(**handler_options)
            elif kwargs['type'] == 'http':
                handler_options['compress'] = kwargs.get('compress', True)
                handler_options['path'] = kwargs.get('path', '/gelf')
                handler_options['timeout'] = kwargs.get('timeout', 5)
                self.handler = GelfHttpHandler(**handler_options)
            else:
                raise Exception(f'Unknown Graylog type "{kwargs["type"]}" for "{kwargs["host"]}:{kwargs["port"]}"')

            self.handler.setFormatter(self._formatter)
        else:
            self._target = target
            # Output produced by remote commands might contain characters which cannot be encoded on Windows deployer.
            # Specify explicitly utf-8 encoding which is native to the remote machines.
            self.handler = FileHandlerWithHeader(self._formatter, self._target, mode=filemode, header=self._header, encoding='utf-8')

        if level not in LOGGING_LEVELS_BY_NAME:
            raise Exception(f'Failed to create logger - unknown logging level: "{level}"')
        self._level = LOGGING_LEVELS_BY_NAME[level]
        self.handler.setLevel(self._level)

    def __str__(self) -> str:
        return f'target: {self._target}, level: {LOGGING_NAMES_BY_LEVEL[self._level]}, colorize: {self._colorize}, datefmt: {self._datefmt}, format: {self._format}'

    def append_to_logger(self, logger: EnhancedLogger) -> None:
        logger.addHandler(self.handler)

    def has_stdout_target(self) -> bool:
        return self._target == 'stdout'


class Log:

    def __init__(self, raw_inventory: dict, handlers: List[LogHandler]):
        logger = logging.getLogger(raw_inventory.get('cluster_name', 'cluster.local'))
        self._logger = cast(EnhancedLogger, logger)
        self._logger.setLevel(VERBOSE)

        if self._logger.hasHandlers():
            self._logger.handlers.clear()

        for handler in handlers:
            handler.append_to_logger(self._logger)

    @property
    def logger(self) -> EnhancedLogger:
        return self._logger


class LoggerWriter:
    def __init__(self, logger: EnhancedLogger, caller: dict, prefix: str) -> None:
        self.logger = logger
        self.caller = caller
        self.prefix = prefix
        self.buf = ""

    def write(self, message: str) -> None:
        # Both remote stderr and stdout are printed to local stdout
        sys.stdout.write(message)

        lines = message.split('\n')
        for line in lines[:-1]:
            self.buf = self.buf + line
            self._log()
        self.buf = self.buf + lines[-1]

    def flush(self, remainder: bool = False) -> None:
        if remainder and self.buf:
            self._log()

    def _log(self) -> None:
        self.logger.log(logging.DEBUG, self.buf, extra={
            'real_caller': self.caller, 'prefix': self.prefix, 'ignore_stdout': True
        })
        self.buf = ""

    def __repr__(self) -> str:
        return f"LoggerWriter{{DEBUG,stdout}} at {hex(id(self))}"


def parse_log_argument(argument: str) -> LogHandler:
    """
    Parse raw CLI arguments and verify for required parameters
    :param argument: Raw CLI argument string. For example: test.log;level=verbose;colorize=true
    :return: Initialized LogHandler
    """
    parameters: Dict[str, Any] = {}
    argument_parts = argument.split(';')
    if not argument_parts:
        raise Exception('Defined logger do not contain parameters')
    target = argument_parts[0]
    level: Optional[str] = None
    for parameter in argument_parts[1:]:
        if parameter == '':
            continue
        value: Union[str, bool, int]
        key, value, *rest = parameter.split('=')
        if key == 'level':
            level = value
        else:
            if key in ['colorize', 'correct_newlines', 'debug', 'compress', 'validate']:
                value = value.lower() in ['true', '1']
            elif key in ['chunk_size', 'timeout', 'port']:
                value = int(value)

            parameters[key] = value

    if level is None:
        raise Exception(f'Logging level is not set for logger "{target}"')
    return LogHandler(target, level, **parameters)


def get_dump_debug_filepath(context: dict) -> Optional[str]:
    args = context['execution_arguments']
    if args['disable_dump']:
        return None

    return os.path.join(args['dump_location'], 'dump', 'debug.log')


def init_log_from_context_args(globals: dict, context: dict, raw_inventory: dict) -> Log:
    """
    Create Log from raw CLI arguments in Cluster context
    :param globals: parsed globals collection
    :param context: context holding execution arguments.
    :param raw_inventory: parsed but not yet enriched inventory
    :return: Initialized Log, based on all parsed logging arguments
    """

    handlers = []
    stdout_specified = False

    args = context['execution_arguments']
    if args.get('log') is not None:
        for argument in args.get('log'):
            handler = parse_log_argument(argument[0])
            if handler.has_stdout_target():
                if stdout_specified:
                    raise Exception('Multiple stdout logs specified')
                else:
                    stdout_specified = True
            handlers.append(handler)

    debug_filepath = get_dump_debug_filepath(context)
    if debug_filepath:
        handlers.append(LogHandler(target=debug_filepath,
                                   **globals['logging']['default_targets']['dump']))

    if not stdout_specified:
        stdout_settings = deepcopy(globals['logging']['default_targets']['stdout'])
        handlers.append(LogHandler(target='stdout', **stdout_settings))

    log = Log(raw_inventory, handlers)

    log.logger.verbose('Using the following loggers: \n\t%s' % "\n\t".join("- " + str(x) for x in handlers))

    return log


def caller_info(logger: EnhancedLogger) -> Dict[str, object]:
    """
    Catches and returns invocation metadata of the method that calls caller_info()

    :param logger: EnhancedLogger
    :return: dictionary with the invocation metadata
    """
    fn, lno, func, sinfo = logger.findCaller()
    record: logging.LogRecord = logger.makeRecord("", logging.DEBUG, fn, lno, "", (), None,
                                                  func=func, extra=None, sinfo=sinfo)
    return dict(item for item in record.__dict__.items()
                if item[0] in (
                    # record's fields describing invocation origin
                    'pathname', 'filename', 'module', 'lineno', 'funcName', 'stack_info',
                    # record's fields describing initial process and thread context
                    'thread', 'threadName', 'process', 'processName'
                ))



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import io
import json
import os
import re
import shutil
import sys
import time
import tarfile

from typing import Tuple, Callable, List, TextIO, cast, Union, TypeVar, Dict

import yaml
import ruamel.yaml
from copy import deepcopy
from datetime import datetime
from collections import OrderedDict

from ruamel.yaml import CommentedMap
from typing_extensions import Protocol

from kubemarine.core import log
from kubemarine.core.errors import pretty_print_error


_T_contra = TypeVar("_T_contra", contravariant=True)


class SupportsDunderLT(Protocol[_T_contra]):
    def __lt__(self, __other: _T_contra) -> bool: ...


def do_fail(message: str = '', reason: Exception = None, hint: str = '', logger: log.EnhancedLogger = None) -> None:

    if not logger:
        sys.stderr.write("\033[91m")

    pretty_print_error(message, reason, logger)

    # Please do not rewrite this to logging approach:
    # hint should be visible only in stdout and without special formatting
    if hint != "":
        sys.stderr.write("\n")
        sys.stderr.write(hint + "\n")

    if not logger:
        sys.stderr.write("\033[0m")

    sys.exit(1)


def get_elapsed_string(start: float, end: float) -> str:
    elapsed = end - start
    hours, remainder = divmod(elapsed, 3600)
    minutes, seconds = divmod(remainder, 60)
    return '{:02}h {:02}m {:02}s'.format(int(hours), int(minutes), int(seconds))


def prepare_dump_directory(location: str, reset_directory: bool = True) -> None:
    dumpdir = os.path.join(location, 'dump')
    if reset_directory and os.path.exists(dumpdir) and os.path.isdir(dumpdir):
        shutil.rmtree(dumpdir)
    os.makedirs(dumpdir, exist_ok=True)


def make_ansible_inventory(location: str, c: object) -> None:
    from kubemarine.core.cluster import KubernetesCluster
    cluster = cast(KubernetesCluster, c)

    inventory = get_final_inventory(cluster)
    roles = []
    for node in inventory['nodes']:
        for role in node['roles']:
            if role not in roles:
                roles.append(role)

    config: dict = {
        'all': [
            'localhost ansible_connection=local'
        ],
        'cluster:children': []
    }

    already_global_defined = []

    for role in roles:
        config[role] = []
        config['cluster:children'].append(role)
        for node in cluster.nodes[role].get_final_nodes().get_ordered_members_configs_list():
            record = "%s ansible_host=%s ansible_ssh_user=%s ansible_ssh_pass=%s ansible_ssh_private_key_file=%s ip=%s" % \
                     (node['name'],
                      node['connect_to'],
                      node.get('username', cluster.globals['connection']['defaults']['username']),
                      node.get('password'),
                      node.get('keyfile'),
                      node['internal_address'])
            if node.get('address') is not None:
                record += ' external_ip=%s' % node['address']

            if node['name'] not in already_global_defined:
                config['all'].append(record)
                # to avoid duplicate definition in global section we have to check is that was already defined?
                already_global_defined.append(node['name'])

            config[role].append(node['name'])

    config['cluster:vars'] = [
        'ansible_become=True'
    ]

    for group in ['services', 'plugins']:
        if inventory.get(group) is not None:
            for service_name, service_configs in inventory[group].items():
                # write to inventory only plugins, which will be installed
                if group != 'plugins' or service_configs.get('install', False):

                    config['cluster:vars'].append('\n# %s.%s' % (group, service_name))

                    if isinstance(service_configs, dict):

                        if service_configs.get('installation') is not None:
                            del service_configs['installation']
                        if service_configs.get('install') is not None:
                            del service_configs['install']

                        for config_name, config_value in service_configs.items():
                            if isinstance(config_value, dict) or isinstance(config_value, list):
                                config_value = json.dumps(config_value)
                            config['cluster:vars'].append('%s_%s=%s' % (
                                # TODO: Rewrite replace using regex
                                service_name.replace('-', '_').replace('.', '_').replace('/', '_'),
                                config_name.replace('-', '_').replace('.', '_').replace('/', '_'),
                                config_value))
                    else:
                        config_value = json.dumps(service_configs)
                        config['cluster:vars'].append('%s=%s' % (
                            service_name.replace('-', '_').replace('.', '_'),
                            config_value))

    config_compiled = ''
    for section_name, strings in config.items():
        config_compiled += '[%s]' % section_name
        for string in strings:
            config_compiled += '\n' + string
        config_compiled += '\n\n'

    dump_file({}, config_compiled, location, dump_location=False)


def get_current_timestamp_formatted() -> str:
    return datetime.now().strftime("%Y%m%d-%H%M%S")


def get_final_inventory(c: object, initial_inventory: dict = None) -> dict:
    from kubemarine.core.cluster import KubernetesCluster
    cluster = cast(KubernetesCluster, c)

    if initial_inventory is None:
        inventory = deepcopy(cluster.inventory)
    else:
        inventory = deepcopy(initial_inventory)

    from kubemarine import admission, kubernetes, packages, plugins, thirdparties
    from kubemarine.plugins import nginx_ingress
    from kubemarine.procedures import add_node, remove_node, migrate_cri

    inventory_finalize_functions = {
        add_node.add_node_finalize_inventory,
        remove_node.remove_node_finalize_inventory,
        kubernetes.upgrade_finalize_inventory,
        thirdparties.upgrade_finalize_inventory,
        plugins.upgrade_finalize_inventory,
        packages.upgrade_finalize_inventory,
        admission.finalize_inventory,
        nginx_ingress.finalize_inventory,
        migrate_cri.migrate_cri_finalize_inventory
    }

    for finalize_fn in inventory_finalize_functions:
        inventory = finalize_fn(cluster, inventory)

    return inventory


def merge_vrrp_ips(procedure_inventory: dict, inventory: dict) -> None:
    if "vrrp_ips" in inventory and len(inventory["vrrp_ips"]) > 0:
        raise Exception("vrrp_ips section already defined, merging not supported yet")
    else:
        inventory["vrrp_ips"] = procedure_inventory["vrrp_ips"]

    if isinstance(inventory, OrderedDict):
        inventory.move_to_end("vrrp_ips", last=False)


def dump_file(context: Union[dict, object], data: Union[TextIO, str], filename: str,
              *, dump_location: bool = True) -> None:
    if dump_location:
        if not isinstance(context, dict):
            # cluster is passed instead of the context directly
            from kubemarine.core.cluster import KubernetesCluster
            cluster = cast(KubernetesCluster, context)
            context = cluster.context

        args = context['execution_arguments']
        if args['disable_dump'] \
                and not (filename in ClusterStorage.PRESERVED_DUMP_FILES and context['preserve_inventory']):
            return

        prepare_dump_directory(args['dump_location'], reset_directory=False)
        target_path = get_dump_filepath(context, filename)
    else:
        target_path = get_external_resource_path(filename)

    if isinstance(data, io.StringIO):
        text = data.getvalue()
    elif isinstance(data, str):
        text = data
    else:
        text = data.read()

    with open_utf8(target_path, 'w') as file:
        file.write(text)


def get_dump_filepath(context: dict, filename: str) -> str:
    if context.get("dump_filename_prefix"):
        filename = f"{context['dump_filename_prefix']}_{filename}"

    return get_external_resource_path(os.path.join(context['execution_arguments']['dump_location'], 'dump', filename))


def wait_command_successful(g: object, command: str,
                            retries: int = 15, timeout: int = 5,
                            warn: bool = True, hide: bool = False) -> None:
    from kubemarine.core.group import NodeGroup
    group = cast(NodeGroup, g)

    log = group.cluster.log

    while retries > 0:
        log.debug("Waiting for command to succeed, %s retries left" % retries)
        result = group.sudo(command, warn=warn, hide=hide)
        exit_code = list(result.values())[0].exited
        if exit_code == 0:
            log.debug("Command succeeded")
            return
        retries = retries - 1
        time.sleep(timeout)
    raise Exception("Command failed")


def open_utf8(path: str, mode: str = 'r') -> TextIO:
    return cast(TextIO, open(path, mode + 't', encoding='utf-8'))


def open_internal(path: str, mode: str = 'r') -> TextIO:
    return open_utf8(get_internal_resource_path(path), mode)


def open_external(path: str, mode: str = 'r') -> TextIO:
    return open_utf8(get_external_resource_path(path), mode)


def read_internal(path: str) -> str:
    with open_internal(path) as f:
        return f.read()


def read_external(path: str) -> str:
    with open_external(path) as f:
        return f.read()


def get_external_resource_path(path: str) -> str:
    return os.path.abspath(path)


def get_internal_resource_path(path: str) -> str:
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..', path)
    )


def determine_resource_absolute_file(path: str) -> Tuple[str, bool]:
    """
    Get and verify absolute path to resource file
    :param path: Relative path to resource
    :return: Tuple of absolute path to resource file and flag defining if is an external resource
    """
    # is resource exists as it is defined?
    initial_definition = get_external_resource_path(path)
    if os.path.isfile(initial_definition):
        return initial_definition, True

    # is resource exists as internal resource?
    patched_definition = get_internal_resource_path(path)
    if os.path.isfile(patched_definition):
        return patched_definition, False

    raise Exception('Requested resource %s is not exists at %s or %s' % (path, initial_definition, patched_definition))


def determine_resource_absolute_dir(path: str) -> Tuple[str, bool]:
    """
    Get and verify absolute path to resource directory
    :param path: Relative path to resource
    :return: Tuple of absolute path to resource directory and flag defining if is an external resource
    """
    dirname = os.path.dirname(path)
    # is resource dir exists as it is defined?
    initial_definition = get_external_resource_path(dirname)
    if os.path.isdir(initial_definition):
        return initial_definition, True

    # is resource dir exists as internal resource?
    patched_definition = get_internal_resource_path(dirname)
    if os.path.isdir(patched_definition):
        return patched_definition, False

    raise Exception(
        'Requested resource directory %s is not exists at %s or %s' % (path, initial_definition, patched_definition))


def get_local_file_sha1(filename: str) -> str:
    sha1 = hashlib.sha1()

    # Read local file by chunks of 2^16 bytes (65536) and calculate aggregated SHA1
    with open(filename, 'rb') as f:
        while True:
            data = f.read(2 ** 16)
            if not data:
                break
            sha1.update(data)

    return sha1.hexdigest()


def yaml_structure_preserver() -> ruamel.yaml.YAML:
    """YAML loader and dumper which saves original structure"""
    ruamel_yaml = ruamel.yaml.YAML()
    ruamel_yaml.preserve_quotes = True
    return ruamel_yaml


def is_sorted(l: list, key: Callable = None) -> bool:
    """
    Check that the specified list is sorted.

    :param l: list to check
    :param key: custom key function to customize the sort order
    :return: boolean flag if the list is sorted
    """
    if key is None:
        key = lambda x: x
    return all(key(l[i]) <= key(l[i + 1]) for i in range(len(l) - 1))


def map_sorted(map_: CommentedMap, key: Callable[[str], SupportsDunderLT] = None) -> CommentedMap:
    """
    Check that the specified CommentedMap is sorted, or create new sorted map from it otherwise.

    :param map_: CommentedMap instance to check
    :param key: custom key function to customize the sort order of the map keys
    :return: the same or new sorted instance of the map
    """
    if key is not None:
        _key = key
    else:
        _key = lambda x: x
    map_keys = list(map_)
    if not is_sorted(map_keys, key=_key):
        map_ = CommentedMap(sorted(map_.items(), key=lambda item: _key(item[0])))

    return map_


def insert_map_sorted(map_: CommentedMap, k: str, v: object, key: Callable = None) -> None:
    """
    Insert new item to the CommentedMap or update the value for the existing key.
    The map should be already sorted.

    :param map_: sorted CommentedMap instance
    :param k: new key
    :param v: new value
    :param key: custom key function to customize the sort order of the map keys
    """
    if k in map_:
        map_[k] = v
        return

    if key is None:
        key = lambda x: x
    # Find position to insert new item maintaining the order
    pos = max((mi + 1 for mi, mv in enumerate(map_)
               if key(mv) < key(k)),
              default=0)

    map_.insert(pos, k, v)


def load_yaml(filepath: str) -> dict:
    try:
        with open_utf8(filepath, 'r') as stream:
            data: dict = yaml.safe_load(stream)
            return data
    except yaml.YAMLError as exc:
        do_fail(f"Failed to load {filepath}", exc)
        return {}  # unreachable


def true_or_false(value: Union[str, bool]) -> str:
    """
    The method check string and boolean value
    :param value: Value that should be checked
    """
    input_string = str(value)
    if input_string in ['true', 'True', 'TRUE']:
        result = "true"
    elif input_string in ['false', 'False', 'FALSE']:
        result = "false"
    else:
        result = "undefined"
    return result


def get_version_filepath() -> str:
    return get_internal_resource_path("version")


def get_version() -> str:
    return read_internal(get_version_filepath()).strip()


def minor_version(version: str) -> str:
    """
    Converts vN.N.N to vN.N
    """
    return 'v' + '.'.join(map(str, _test_version(version, 3)[0:2]))


def version_key(version: str) -> Tuple[int, int, int]:
    """
    Converts vN.N.N to (N, N, N) that can be used in comparisons.
    """
    v = _test_version(version, 3)
    return v[0], v[1], v[2]


def minor_version_key(version: str) -> Tuple[int, int]:
    """
    Converts vN.N to (N, N) that can be used in comparisons.
    """
    v = _test_version(version, 2)
    return v[0], v[1]


def _test_version(version: str, numbers_amount: int) -> List[int]:
    # catch version without "v" at the first symbol
    if version.startswith('v'):
        version_list: list = version[1:].split('.')
        # catch invalid version 'v1.16'
        if len(version_list) == numbers_amount:
            # parse str to int and catch invalid symbols in version number
            try:
                for i, value in enumerate(version_list):
                    # whitespace required because python's int() ignores them
                    version_list[i] = int(value.replace(' ', '.'))
            except ValueError:
                pass
            else:
                return version_list

    expected_pattern = 'v' + '.'.join('N+' for _ in range(numbers_amount))
    raise ValueError(f'Incorrect version \"{version}\" format, expected version pattern is \"{expected_pattern}\"')


def parse_aligned_table(table_text: str) -> List[Dict[str, str]]:
    """
    Parse aligned table into the list of {header: cell} map.
    The text with table can be initially produced, for example, by https://pkg.go.dev/text/tabwriter.

    :param table_text: aligned table as string
    :return: List of rows
    """
    rows = table_text.strip().split('\n')

    # Parse headers and their positions. No leading or trailing spaces are expected in the line.
    headers_line = rows[0]
    headers = []
    headers_pos = []
    pos = 0
    for match in re.finditer(r'\s+', headers_line):
        headers_pos.append(pos)
        span = match.span()
        headers.append(headers_line[pos:span[0]])
        pos = span[1]

    headers_pos.append(pos)
    headers.append(headers_line[pos:len(headers_line)])

    # Parse each row in the table,
    # provided that the columns start at the same positions as the headers
    headers_num = len(headers)
    data = []
    for row_text in rows[1:]:
        row = {}
        for i, header in enumerate(headers):
            if i == headers_num - 1:
                cell = row_text[headers_pos[i]:]
            else:
                cell = row_text[headers_pos[i]:headers_pos[i+1]]

            row[header] = cell.strip()

        data.append(row)

    return data


class ClusterStorage:
    """
    File preservation:
    1- Create folder where dumps are stored
    2- Rotating dumps in the storage folder
    3- Uploading dumps to nodes
    4- Copying dumps to new nodes
    """

    PRESERVED_DUMP_FILES = ['procedure.yaml', 'procedure_parameters', 'cluster_precompiled.yaml',
                            'cluster.yaml','cluster_initial.yaml', 'cluster_finalized.yaml']

    def __init__(self, cluster: object):
        from kubemarine.core.cluster import KubernetesCluster
        self.cluster = cast(KubernetesCluster, cluster)
        self.dir_path = "/etc/kubemarine/procedures/"
        self.dir_name = ''
        self.dir_location = ''

    def make_dir(self) -> None:
        """
        This method creates a directory in which logs about operations on the cluster will be stored.
        """
        readable_timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        initial_procedure = self.cluster.context["initial_procedure"]
        self.dir_name = readable_timestamp + "_" + initial_procedure + "/"
        self.dir_location = self.dir_path + self.dir_name
        self.cluster.nodes['control-plane'].get_final_nodes().sudo(f"mkdir -p {self.dir_location} ; sudo rm {self.dir_path + 'latest_dump'} ;"
                                                 f" sudo ln -s {self.dir_location} {self.dir_path + 'latest_dump'}")

    def rotation_file(self) -> None:
        """
        This method packs files with logs and maintains a structured storage of logs on the cluster.
        """
        not_pack_file = self.cluster.inventory['procedure_history']['archive_threshold']
        delete_old = self.cluster.inventory['procedure_history']['delete_threshold']

        command = f'ls {self.dir_path} | grep -v latest_dump'
        node_group_results = self.cluster.nodes["control-plane"].get_final_nodes().sudo(command)
        with node_group_results.get_group().new_executor() as exe:
            for control_plane in exe.group.get_ordered_members_list():
                result = node_group_results[control_plane.get_host()]
                files = result.stdout.split()
                files.sort(reverse=True)
                for i, file in enumerate(files):
                    if i >= not_pack_file and i < delete_old:
                        if 'tar.gz' not in file:
                            control_plane.sudo(
                                f'tar -czvf {self.dir_path + file + ".tar.gz"} {self.dir_path + file} &&'
                                f'sudo rm -r {self.dir_path + file}')
                    elif i >= delete_old:
                        control_plane.sudo(f'rm -rf {self.dir_path + file}')

    def compress_and_upload_archive(self) -> None:
        """
        This method compose dump files and sends the collected files to the nodes.
        """
        context = self.cluster.context
        archive = get_dump_filepath(context, "local.tar.gz")
        with tarfile.open(archive, "w:gz") as tar:
            for name in ClusterStorage.PRESERVED_DUMP_FILES:
                source = get_dump_filepath(context, name)
                if os.path.exists(source):
                    tar.add(source, 'dump/' + name)
            tar.add(context['execution_arguments']['config'], 'cluster.yaml')
            tar.add(get_version_filepath(), 'version')

        self.cluster.log.debug('Uploading archive with preserved information about the procedure.')
        self.cluster.nodes['control-plane'].get_final_nodes().put(archive, self.dir_location + 'local.tar.gz', sudo=True)
        self.cluster.nodes['control-plane'].get_final_nodes().sudo(f'tar -C {self.dir_location} -xzv --no-same-owner -f {self.dir_location + "local.tar.gz"}  && '
                                                 f'sudo rm -f {self.dir_location + "local.tar.gz"} ')

    def collect_procedure_info(self) -> None:
        """
        This method collects information about the type of procedure and the version of the tool we are working with.
        """
        context = self.cluster.context
        out = dict()
        out['arguments'] = context['initial_cli_arguments']
        if 'proceeded_tasks' in context:
            out['finished_tasks'] = context['proceeded_tasks']
        out["initial_procedure"] = context["initial_procedure"]
        out["successfully_performed"] = context["successfully_performed"]
        out['status'] = context['status']
        output = yaml.dump(out)
        dump_file(context, output, "procedure_parameters")

    def upload_info_new_control_planes(self) -> None:
        """
        This method is used to transfer backup logs from the initial control-plane to the new control-planes.
        """
        new_control_planes = self.cluster.nodes['control-plane'].get_new_nodes()
        if new_control_planes.is_empty():
            return

        archive_name = 'dump_log_cluster.tar.gz'
        archive_dump_path = get_dump_filepath(self.cluster.context, archive_name)
        archive_remote_path = f"/tmp/{archive_name}"
        log = self.cluster.log

        control_plane = self.cluster.nodes['control-plane'].get_initial_nodes().get_first_member()
        data_copy_res = control_plane.sudo(f'tar -czvf {archive_remote_path} {self.dir_path}')
        log.verbose("Archive with procedures history is created:\n%s" % data_copy_res)
        control_plane.get(archive_remote_path, archive_dump_path)

        log.debug("Archive with procedures history is downloaded")

        for group in new_control_planes.get_ordered_members_list():
            group.put(archive_dump_path, archive_remote_path, sudo=True)
            group.sudo(f'tar -C / -xzvf {archive_remote_path}')
            log.debug(f"Archive with procedures history is uploaded to {group.get_node_name()!r}")



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from io import StringIO

from kubemarine import system, packages
from kubemarine.core import utils
from kubemarine.core.group import NodeGroup, RunnersGroupResult, CollectorCallback


def install(group: NodeGroup) -> RunnersGroupResult:
    cluster = group.cluster
    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            os_specific_associations = cluster.get_associations_for_node(node.get_host(), 'docker')
            packages.install(node, include=os_specific_associations['package_name'], callback=collector)

            system.enable_service(node, name=os_specific_associations['service_name'],
                                  now=True, callback=collector)

            # remove previous daemon.json to avoid problems in case when previous config was broken
            node.sudo("rm -f %s && sudo systemctl restart %s"
                      % (os_specific_associations['config_location'],
                         os_specific_associations['service_name']),
                      callback=collector)
    return collector.result


def uninstall(group: NodeGroup) -> RunnersGroupResult:
    # delete all known docker packages
    return packages.remove(group, include=['docker', 'docker-engine', 'docker.io', 'docker-ce'])


def disable(group: NodeGroup) -> None:
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            service_name = exe.cluster.get_package_association_for_node(
                node.get_host(), 'docker', 'service_name')
            system.disable_service(node, name=service_name, now=True)


def configure(group: NodeGroup) -> RunnersGroupResult:
    cluster = group.cluster
    log = cluster.log

    settings_json = json.dumps(cluster.inventory["services"]['cri']['dockerConfig'], sort_keys=True, indent=4)
    utils.dump_file(cluster, settings_json, 'docker-daemon.json')

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            os_specific_associations = exe.cluster.get_associations_for_node(node.get_host(), 'docker')
            log.debug("Uploading docker configuration to %s node..." % node.get_node_name())
            node.put(StringIO(settings_json), os_specific_associations['config_location'],
                     backup=True, sudo=True)
            log.debug("Restarting Docker on %s node..." % node.get_node_name())
            node.sudo(
                f"chmod 600 {os_specific_associations['config_location']} && "
                f"sudo systemctl restart {os_specific_associations['service_name']} && "
                f"sudo {os_specific_associations['executable_name']} info",
                callback=collector)

    return collector.result


def prune(group: NodeGroup) -> RunnersGroupResult:
    return group.sudo('docker container stop $(sudo docker container ls -aq); '
                      'sudo docker container rm $(sudo docker container ls -aq); '
                      'sudo docker system prune -a -f; '
                      # kill all containerd-shim processes, so that no orphan containers remain 
                      'sudo pkill -9 -f "^containerd-shim"', warn=True)



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
import gzip
import io
import json
import os
import shutil
import tarfile
import threading
import time
import uuid
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from typing import List, Tuple, Union, Dict, Optional, Iterator
from typing_extensions import Literal

import yaml

from kubemarine.core import utils, flow, log
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup, RemoteExecutor
from kubemarine.core.resources import DynamicResources


def get_default_backup_files_list(cluster: KubernetesCluster) -> List[str]:
    haproxy_service = cluster.get_package_association('haproxy', 'service_name')
    keepalived_service = cluster.get_package_association('keepalived', 'service_name')

    backup_files_list = [
        "/etc/resolv.conf",
        "/etc/hosts",
        "/etc/chrony.conf",
        "/etc/selinux/config",
        "/etc/yum.repos.d/",
        "/var/lib/kubelet/pki",
        "/etc/modules-load.d/",
        "/etc/audit/rules.d/",
        "/etc/haproxy/haproxy.cfg",
        f"/etc/systemd/system/{haproxy_service}.service.d/{haproxy_service}.conf",
        "/etc/keepalived/keepalived.conf",
        f"/etc/systemd/system/{keepalived_service}.service.d/{keepalived_service}.conf",
        "/usr/local/bin/check_haproxy.sh",
        "/etc/kubernetes",
        "/root/.kube/config",
        "/etc/systemd/system/kubelet.service"
    ]

    cri_impl = cluster.inventory['services']['cri']['containerRuntime']
    if cri_impl == "docker":
        backup_files_list.append("/etc/docker/daemon.json")
    elif cri_impl == "containerd":
        backup_files_list.append("/etc/containerd/config.toml")
        backup_files_list.append("/etc/crictl.yaml")

    return backup_files_list


def prepare_backup_tmpdir(cluster: KubernetesCluster) -> str:
    backup_directory = cluster.context.get('backup_tmpdir')
    if not backup_directory:
        cluster.log.verbose('Backup directory is not ready yet, preparing..')
        backup_directory = cluster.context['backup_tmpdir'] = utils.get_dump_filepath(cluster.context, 'backup')
        shutil.rmtree(backup_directory, ignore_errors=True)
        os.mkdir(backup_directory)
        cluster.log.verbose('Backup directory prepared')
    return backup_directory


def verify_backup_location(cluster: KubernetesCluster) -> None:
    target = utils.get_external_resource_path(cluster.procedure_inventory.get('backup_location', 'backup.tar.gz'))
    if not os.path.isdir(target) and not os.path.isdir(os.path.abspath(os.path.join(target, os.pardir))):
        raise FileNotFoundError('Backup location directory not exists')


def export_ansible_inventory(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster)
    shutil.copyfile(cluster.context['execution_arguments']['ansible_inventory_location'],
                    os.path.join(backup_directory, 'ansible-inventory.ini'))
    cluster.log.verbose('ansible-inventory.ini exported to backup')


def export_packages_list(cluster: KubernetesCluster) -> None:
    cluster.context['backup_descriptor']['nodes']['packages'] = {}
    if cluster.get_os_family() in ['rhel', 'rhel8']:
        cmd = r"rpm -qa"
    else:
        cmd = r"dpkg-query -f '${Package}=${Version}\n' -W"
    results = cluster.nodes['all'].sudo(cmd)
    for host, result in results.items():
        cluster.context['backup_descriptor']['nodes']['packages'][host] = result.stdout.strip().split('\n')


def export_hostname(cluster: KubernetesCluster) -> None:
    cluster.context['backup_descriptor']['nodes']['hostnames'] = {}
    results = cluster.nodes['all'].sudo('hostnamectl status | head -n 1 | sed -e \'s/[a-zA-Z ]*://g\'')
    cluster.log.verbose(results)
    for host, result in results.items():
        cluster.context['backup_descriptor']['nodes']['hostnames'][host] = result.stdout.strip()


def export_cluster_yaml(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster)
    shutil.copyfile(utils.get_dump_filepath(cluster.context, 'cluster.yaml'),
                    os.path.join(backup_directory, 'cluster.yaml'))
    shutil.copyfile(utils.get_external_resource_path(cluster.context['execution_arguments']['config']),
                    os.path.join(backup_directory, 'original_cluster.yaml'))
    cluster.log.verbose('cluster.yaml exported to backup')


def export_nodes(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster)
    backup_nodes_data_dir = os.path.join(backup_directory, 'nodes_data')
    os.mkdir(backup_nodes_data_dir)

    backup_list = get_default_backup_files_list(cluster)
    for filepath, enabled in cluster.procedure_inventory.get('backup_plan', {}).get('nodes', {}).items():
        if not enabled and filepath in backup_list:
            backup_list.remove(filepath)
        if enabled and filepath not in backup_list:
            backup_list.append(filepath)

    cluster.log.debug('Backing up the following files: \n' + '  - ' + '\n  - '.join(backup_list))

    backup_command = 'cd /tmp && ' \
                     'sudo tar -czvf /tmp/kubemarine-backup.tar.gz -P $(sudo readlink -e %s) && ' \
                     'sudo ls -la /tmp/kubemarine-backup.tar.gz && ' \
                     'sudo du -hs /tmp/kubemarine-backup.tar.gz' % (' '.join(backup_list))

    data_copy_res = cluster.nodes['all'].run(backup_command)

    cluster.log.debug('Backup created:\n%s' % data_copy_res)

    cluster.log.debug('Downloading nodes backups:')
    for node in cluster.nodes['all'].get_ordered_members_list():
        node.get('/tmp/kubemarine-backup.tar.gz',
                 os.path.join(backup_nodes_data_dir, '%s.tar.gz' % node.get_node_name()))
        cluster.log.debug('Backup \'%s\' downloaded' % node.get_node_name())

    cluster.log.verbose('Deleting backup file from nodes...')
    cluster.nodes['all'].sudo('rm -f /tmp/kubemarine-backup.tar.gz')


def export_etcd(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster)
    etcd_node, is_custom_etcd_node = select_etcd_node(cluster)
    cluster.context['backup_descriptor']['etcd']['image'] = retrieve_etcd_image(cluster, etcd_node)

    # Try to detect cluster health and other metadata like db size, leader
    etcd_status = None
    try:
        status_json = etcd_node.sudo('etcdctl endpoint status --cluster -w json').get_simple_out()
        cluster.log.verbose(status_json)

        etcd_status = json.load(io.StringIO(status_json.lower()))
        parsed_etcd_status = {}
        for item in etcd_status:
            # get rid of https:// and :2379
            address = item['endpoint'].split('//')[1].split(':')[0]
            node_name = cluster.nodes['all'].get_first_member(apply_filter={"internal_address": address}).get_node_name()
            parsed_etcd_status[node_name] = item
        cluster.context['backup_descriptor']['etcd']['status'] = parsed_etcd_status
    except Exception:
        cluster.log.verbose('Failed to load and parse ETCD status')

    if is_custom_etcd_node:
        cluster.context['backup_descriptor']['etcd']['source'] = etcd_node.get_node_name()
    else:
        # if user did not provide node, then we have to select leader by ourselves
        if not etcd_status:
            raise Exception('Failed to load ETCD status and impossible to detect ETCD leader for making snapshot from it')
        etcd_leader_id = etcd_status[0]['status']['leader']
        etcd_leader_name = None
        for name, item in parsed_etcd_status.items():
            if item['status']['header']['member_id'] == etcd_leader_id:
                etcd_leader_name = name
        if etcd_leader_name:
            cluster.log.verbose('Detected ETCD leader: %s' % etcd_leader_name)
            cluster.context['backup_descriptor']['etcd']['source'] = etcd_leader_name
            etcd_node = cluster.nodes['control-plane'].get_member_by_name(etcd_leader_name)
        else:
            raise Exception('Failed to detect ETCD leader - not possible to create backup from actual DB')

    snap_name = 'snapshot%s.db' % int(round(time.time() * 1000))
    endpoint_ip = etcd_node.get_config()["internal_address"]
    cluster.log.debug('Creating ETCD backup "%s"...' % snap_name)
    result = etcd_node.sudo(f'etcdctl snapshot save /var/lib/etcd/{snap_name} --endpoints=https://{endpoint_ip}:2379 '
                            f'&& sudo mv /var/lib/etcd/{snap_name} /tmp/{snap_name} '
                            f'&& sudo ls -la /tmp/{snap_name} '
                            f'&& sudo du -hs /tmp/{snap_name} '
                            f'&& sudo chmod 666 /tmp/{snap_name}', timeout=600)
    cluster.log.debug(result)
    etcd_node.get('/tmp/' + snap_name, backup_directory + '/etcd.db')
    cluster.log.verbose('Deleting ETCD snapshot file from "%s"...')
    etcd_node.sudo('rm -f /tmp/%s' % snap_name)


def select_etcd_node(cluster: KubernetesCluster) -> Tuple[NodeGroup, bool]:
    custom_etcd_node = cluster.procedure_inventory.get('backup_plan', {}).get('etcd', {}).get('source_node')

    if custom_etcd_node:
        if not cluster.nodes['all'].has_node(custom_etcd_node):
            raise Exception('Unknown ETCD node selected as source')
        etcd_node = cluster.nodes['all'].get_member_by_name(custom_etcd_node)
        return etcd_node, True
    else:
        return cluster.nodes['control-plane'].get_any_member(), False


def retrieve_etcd_image(cluster: KubernetesCluster, etcd_node: NodeGroup) -> str:
    # TODO: Detect ETCD version via /etc/kubernetes/manifests/etcd.yaml config if presented, otherwise use containers
    node_name = etcd_node.get_node_name()
    if "docker" == cluster.inventory['services']['cri']['containerRuntime']:
        cont_inspect = "docker inspect $(sudo docker ps -a | grep etcd-%s | awk '{print $1; exit}')" % node_name
        etcd_container_json = json.loads(list(etcd_node.sudo(cont_inspect).values())[0].stdout)[0]
        etcd_image_sha = etcd_container_json['Image'][7:]  # remove "sha256:" prefix

        images_result = etcd_node.sudo("docker image ls --format '{{json .}}'")
        formatted_images_result = "[" + ",".join(list(images_result.values())[0].stdout.strip().split('\n')) + "]"
        images_json = json.loads(formatted_images_result)
        for image in images_json:
            if image['ID'] == etcd_image_sha[:len(image['ID'])]:
                return f"{image['Repository']}:{image['Tag']}"
    else:
        cont_search = "sudo crictl ps --label io.kubernetes.pod.name=etcd-%s -aq | awk '{print $1; exit}'" % node_name
        cont_inspect = f"crictl inspect $({cont_search})"
        etcd_container_json = json.loads(list(etcd_node.sudo(cont_inspect).values())[0].stdout)
        etcd_image_sha = etcd_container_json['info']['config']['image']['image']

        images_json = json.loads(list(etcd_node.sudo("crictl images -v -o json").values())[0].stdout)['images']
        for image in images_json:
            if image['id'] == etcd_image_sha:
                return f"{image['repoTags'][0]}"

    raise Exception("Unable to find etcd image on node %s" % node_name)


def export_kubernetes_version(cluster: KubernetesCluster) -> None:
    control_plane = cluster.nodes['control-plane'].get_any_member()
    version = control_plane.sudo('kubectl get nodes --no-headers | head -n 1 | awk \'{print $5; exit}\'').get_simple_out()
    cluster.context['backup_descriptor']['kubernetes']['version'] = version.strip()


def _load_namespaces(logger: log.EnhancedLogger, control_plane: NodeGroup) -> List[Dict[str, str]]:
    namespaces_result = control_plane.sudo(
        'kubectl get ns -o jsonpath=\'{range .items[*]}{.metadata.name}{"\\n"}{end}\'')
    logger.verbose(namespaces_result)
    parsed_namespaces = namespaces_result.get_simple_out().strip().split('\n')
    return [{'name': name} for name in parsed_namespaces]


def _load_resources(logger: log.EnhancedLogger, control_plane: NodeGroup, namespaced: bool) -> List[Dict[str, str]]:
    resources_result = control_plane.sudo(f'kubectl api-resources --verbs=list --sort-by=name '
                                          f'--namespaced{"" if namespaced else "=false"}')
    logger.verbose(resources_result)
    resources_table = utils.parse_aligned_table(resources_result.get_simple_out())

    resources = [
        {
            'name': _resolve_full_resource_name(row['NAME'], row['APIVERSION']),
            'kind': row['KIND'],
            'apiVersion': row['APIVERSION'],
        }
        for row in resources_table
    ]

    return [r for r in resources if r['name'] not in ('events.events.k8s.io', 'events')]


def _resolve_full_resource_name(name: str, apiversion: str) -> str:
    if '/' not in apiversion:
        return name

    return name + '.' + apiversion[0:apiversion.rindex('/')]


def _filter_resources_by_proposed(logger: log.EnhancedLogger,
                                  loaded_resources: List[Dict[str, str]],
                                  proposed_resources: Union[List[str], Literal['all']],
                                  kind: str) -> List[Dict[str, str]]:
    resources = []
    for resource in loaded_resources:
        name = resource['name']
        if proposed_resources == 'all' or name in proposed_resources:
            resources.append(resource)
        else:
            logger.verbose(f'{kind.capitalize()} "{name}" excluded')

    if proposed_resources != 'all':
        for proposed_resource in proposed_resources:
            if not any(proposed_resource == resource['name'] for resource in resources):
                raise Exception(f'Proposed {kind} "{proposed_resource}" not found in loaded cluster {kind}s')

    logger.debug([r['name'] for r in resources])
    return resources


class ParserPayload:
    def __init__(self, event: str, resource_path: str, namespace: Optional[str]):
        self.event = event
        self.resource_path = resource_path
        self.namespace = namespace


class ExportKubernetesParser:
    def __init__(self,
                 logger: log.EnhancedLogger,
                 backup_directory: str,
                 namespaced_resources: List[Dict[str, str]],
                 nonnamespaced_resources: List[Dict[str, str]],
                 graceful_finish_barrier: int,
                 ):
        self.logger = logger
        self.backup_directory = backup_directory
        self.namespaced_resources = namespaced_resources
        self.nonnamespaced_resources = nonnamespaced_resources
        self.namespaced_resources_result_map: Dict[str, List[str]] = {}
        self.nonnamespaced_resources_result: List[str] = []
        self.closed = False
        self.elapsed: float = 0
        self.total_files = 0
        self._graceful_finish_barrier = graceful_finish_barrier
        self._task_queue: 'Queue[ParserPayload]' = Queue()
        self._lock = threading.Lock()

    def schedule(self, payload: ParserPayload) -> None:
        with self._lock:
            if not self.closed:
                self._task_queue.put(payload)
            else:
                self._clear(payload)

    def finish(self, graceful: bool) -> None:
        with self._lock:
            self._graceful_finish_barrier -= 1
            if not self.closed and (self._graceful_finish_barrier == 0 or not graceful):
                self._task_queue.put(ParserPayload("end", "", None))
                self._graceful_finish_barrier = 0

    def run(self) -> None:
        try:
            while True:
                payload = self._task_queue.get()
                if payload.event == 'end':
                    return

                start = time.time()
                self._handle(payload)
                # do not clear the currently processing payload in case of exception
                self._clear(payload)
                self.elapsed += (time.time() - start)
        finally:
            with self._lock:
                self.closed = True
                # Queue might be not empty in case of exception. Need to clear all pending payloads
                for payload in self._task_queue.queue:
                    self._clear(payload)

    @staticmethod
    def _clear(payload: ParserPayload) -> None:
        if payload.event == 'do':
            os.remove(payload.resource_path)

    def _parse_identity(self, line: str, api_version: str, kind: str) -> Tuple[str, str]:
        if kind == '':
            kind = self._parse_single_line_property(line, 'kind')
        if api_version == '':
            api_version = self._parse_single_line_property(line, 'apiVersion')

        return api_version, kind

    @staticmethod
    def _parse_single_line_property(line: str, key: str) -> str:
        item_prop = line[2:]
        val = ''
        if item_prop.startswith(f'{key}:'):
            val = yaml.safe_load(item_prop)[key]

        return val

    def _handle(self, payload: ParserPayload) -> None:
        namespace = payload.namespace
        if namespace:
            self.logger.debug(f"Loading resources from namespace {namespace!r}...")
        else:
            self.logger.debug(f"Loading non-namespaced resources...")

        resources = self.nonnamespaced_resources if namespace is None else self.namespaced_resources
        items_by_resource: Dict[str, List[str]] = {}

        def append_item(api_version: str, kind: str, item: str) -> None:
            resource_name = next(r['name'] for r in resources if r['apiVersion'] == api_version and r['kind'] == kind)
            items_by_resource.setdefault(resource_name, []).append(item)

        with gzip.open(payload.resource_path, 'rt', encoding='utf-8') as file:
            header = ''
            remainder = ''
            item = ''
            api_version, kind = '', ''
            stage = 'header'
            for line in file:
                if stage == 'header':
                    if line.startswith('- '):
                        stage = 'items'
                        item = line
                        api_version, kind = self._parse_identity(line, '', '')
                    else:
                        header += line
                elif stage == 'items':
                    if line.startswith('- '):
                        append_item(api_version, kind, item)
                        item = line
                        api_version, kind = self._parse_identity(line, '', '')
                    elif not line[0].isspace():
                        stage = 'remainder'
                        append_item(api_version, kind, item)
                        remainder = line
                    else:
                        item += line
                        api_version, kind = self._parse_identity(line, api_version, kind)
                elif stage == 'remainder':
                    remainder += line

        for resource in resources:
            resource_name = resource['name']
            items = items_by_resource.get(resource_name)
            if items is None:
                continue

            location = os.path.join(self.backup_directory, 'kubernetes_resources')
            if namespace is not None:
                location = os.path.join(location, namespace)

            os.makedirs(location, exist_ok=True)

            resource_file_path = os.path.join(location, '%s.yaml' % resource_name)
            self.logger.verbose(
                f"Dumping list of resource {resource_name!r}"
                f"{'' if namespace is None else (' for namespace ' + repr(namespace))}...")
            with utils.open_utf8(resource_file_path, 'w') as file:
                file.write(header)
                for item in items:
                    file.write(item)
                file.write(remainder)

            if namespace is None:
                self.nonnamespaced_resources_result.append(resource_name)
            else:
                self.namespaced_resources_result_map.setdefault(namespace, []).append(resource_name)

            self.total_files += 1


class DownloaderPayload:
    def __init__(self, namespace: Optional[str], resources: List[str]):
        self.namespace = namespace
        self.resources = resources


class DownloaderTasksQueue(Iterator[DownloaderPayload]):
    def __init__(self,
                 namespaces: List[Dict[str, str]],
                 namespaced_resources: List[Dict[str, str]],
                 nonnamespaced_resources: List[Dict[str, str]]
                 ):
        self.namespaces = namespaces
        self.namespaced_resources = namespaced_resources
        self.nonnamespaced_resources = nonnamespaced_resources
        self._tasks = self._unsafe_tasks()
        self._lock = threading.Lock()

    def __next__(self) -> DownloaderPayload:
        with self._lock:
            return next(self._tasks)

    def _unsafe_tasks(self) -> Iterator[DownloaderPayload]:
        yield DownloaderPayload(None, [r['name'] for r in self.nonnamespaced_resources])

        for namespace in self.namespaces:
            yield DownloaderPayload(namespace['name'], [r['name'] for r in self.namespaced_resources])


class DownloadException(Exception):
    def __init__(self, task: DownloaderPayload, reason: BaseException):
        self.task = task
        self.reason = reason


class ExportKubernetesDownloader:
    def __init__(self,
                 backup_directory: str,
                 control_plane: NodeGroup,
                 cluster: KubernetesCluster,
                 tasks_queue: Iterator[DownloaderPayload],
                 parser: ExportKubernetesParser,
                 ):
        self.backup_directory = backup_directory
        self.control_plane = control_plane
        self.connection_pool = cluster.create_connection_pool(control_plane.get_hosts())
        self.tasks_queue = tasks_queue
        self.parser = parser
        self.elapsed: float = 0

    def run(self) -> None:
        start = time.time()
        task: Optional[DownloaderPayload] = None
        try:
            while True:
                if self.parser.closed:
                    # No need to continue as the parser is already closed due to error.
                    break

                task = next(self.tasks_queue, None)
                if task is None:
                    break

                random = uuid.uuid4().hex
                temp_local_filepath = os.path.join(self.backup_directory, random)
                self._download(task, temp_local_filepath)

                self.parser.schedule(ParserPayload("do", temp_local_filepath, task.namespace))
        except BaseException as e:
            self.parser.finish(graceful=False)
            if task is not None:
                raise DownloadException(task, e)
            else:
                raise
        else:
            self.parser.finish(graceful=True)
        finally:
            self.elapsed = time.time() - start
            self.connection_pool.close()

    def _download(self, task: DownloaderPayload, temp_local_filepath: str) -> None:
        namespace = task.namespace
        temp_remote_filepath = f"/tmp/{os.path.basename(temp_local_filepath)}"

        cmd = f'(set -o pipefail && sudo kubectl ' \
              f'{"" if namespace is None else ("-n " + namespace + " ")}' \
              f'get --ignore-not-found ' \
              f'{",".join(r for r in task.resources)} ' \
              f'-o yaml | gzip -c) > {temp_remote_filepath}'

        # Use own connection instance to run commands
        with RemoteExecutor(self.control_plane, self.connection_pool) as exe:
            group = exe.group
            group.run(cmd)
            group.get(temp_remote_filepath, temp_local_filepath)
            group.sudo(f'rm {temp_remote_filepath}')


def export_kubernetes(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster)
    control_plane = cluster.nodes['control-plane'].get_any_member()
    backup_kubernetes = cluster.procedure_inventory.get('backup_plan', {}).get('kubernetes', {})
    logger = cluster.log

    logger.debug('Loading namespaces:')
    loaded_namespaces = _load_namespaces(logger, control_plane)
    proposed_namespaces = backup_kubernetes.get('namespaced_resources', {}).get('namespaces', 'all')
    namespaces = _filter_resources_by_proposed(logger, loaded_namespaces, proposed_namespaces, 'namespace')

    logger.debug('Loading namespaced resource types:')
    loaded_resources = _load_resources(logger, control_plane, True)
    proposed_resources = backup_kubernetes.get('namespaced_resources', {}).get('resources', 'all')
    namespaced_resources = _filter_resources_by_proposed(logger, loaded_resources, proposed_resources, 'resource')

    logger.debug('Loading non-namespaced resource types:')
    loaded_resources = _load_resources(logger, control_plane, False)
    proposed_resources = backup_kubernetes.get('nonnamespaced_resources', 'all')
    nonnamespaced_resources = _filter_resources_by_proposed(logger, loaded_resources, proposed_resources, 'resource')

    logger.debug('Loading resources:')
    start = time.time()

    # Processing of the particular resource includes
    # 1. downloading of the `kubectl` result (mostly IO)
    # 2. parsing and splitting it into files (can potentially consume CPU)
    control_planes = cluster.nodes['control-plane']
    downloaders_per_control_plane = 2

    parser = ExportKubernetesParser(logger, backup_directory, namespaced_resources, nonnamespaced_resources,
                                    control_planes.nodes_amount() * downloaders_per_control_plane)
    downloader_queue = DownloaderTasksQueue(namespaces, namespaced_resources, nonnamespaced_resources)
    downloaders = [
        ExportKubernetesDownloader(backup_directory, control_plane, cluster,
                                   downloader_queue, parser)
        for _ in range(downloaders_per_control_plane)
        for control_plane in control_planes.get_ordered_members_list()
    ]

    logger.debug(f'Using {len(downloaders)} workers to download resources.')

    with ThreadPoolExecutor(max_workers=len(downloaders) + 1) as tpe:
        downloaders_async = [tpe.submit(downloader.run) for downloader in downloaders]
        parser_async = tpe.submit(parser.run)

        def _graceful_shutdown_downloaders() -> Optional[BaseException]:
            # Although parser may exit successfully, not all resources might be processed due to errors in downloaders.
            # Choose any error in such case.
            exc: Optional[BaseException] = None
            for downloader_async in downloaders_async:
                try:
                    downloader_async.result()
                except BaseException as e:
                    logger.verbose(e)
                    if isinstance(e, DownloadException):
                        logger.error(f"Failed to download resources {','.join(e.task.resources)} for namespace {e.task.namespace}")
                        exc = e.reason
                    else:
                        exc = e

            return exc

        try:
            # Wait for successful parsing of resources or till exception.
            parser_async.result()
        except BaseException:
            # Wait for graceful exiting of downloaders after their currently running command is finished.
            # If few background tasks fail, the parser exception has priority, and the other is logged.
            _graceful_shutdown_downloaders()
            raise

        downloader_exception = _graceful_shutdown_downloaders()
        if downloader_exception is not None:
            raise downloader_exception

    downloaded_resources_descriptor = cluster.context['backup_descriptor']['kubernetes']['resources']
    if parser.namespaced_resources_result_map:
        downloaded_resources_descriptor['namespaced'] = parser.namespaced_resources_result_map

    if parser.nonnamespaced_resources_result:
        downloaded_resources_descriptor['nonnamespaced'] = parser.nonnamespaced_resources_result

    logger.verbose(f'Downloading elapsed: {max(downloader.elapsed for downloader in downloaders)}')
    logger.verbose(f'Parsing elapsed: {parser.elapsed}')
    logger.verbose(f'Total elapsed: {time.time() - start}')
    logger.verbose(f'Total files saved: {parser.total_files}')


def make_descriptor(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster)

    cluster.context['backup_descriptor']['kubernetes']['thirdparties'] = cluster.inventory['services']['thirdparties']
    cluster.context['backup_descriptor']['meta']['time']['finished'] = datetime.datetime.now()

    with utils.open_external(os.path.join(backup_directory, 'descriptor.yaml'), 'w') as output:
        output.write(yaml.dump(cluster.context['backup_descriptor']))


def pack_data(cluster: KubernetesCluster) -> None:
    cluster_name = cluster.inventory['cluster_name']
    backup_directory = prepare_backup_tmpdir(cluster)

    backup_filename = 'backup-%s-%s.tar.gz' % (cluster_name, utils.get_current_timestamp_formatted())

    target = utils.get_external_resource_path(cluster.procedure_inventory.get('backup_location', backup_filename))
    if os.path.isdir(target):
        target = os.path.join(target, backup_filename)

    cluster.log.debug('Packing all data...')
    with tarfile.open(target, "w:gz") as tar_handle:
        for root, dirs, files in os.walk(backup_directory):
            for file in files:
                pathname = os.path.join(root, file)
                tar_handle.add(pathname, pathname.replace(backup_directory, ''))
        tar_handle.close()

    cluster.log.verbose('Cleaning up...')
    shutil.rmtree(backup_directory, ignore_errors=True)


tasks = OrderedDict({
    "verify_backup_location": verify_backup_location,
    "export": {
        "inventory": {
            "cluster_yaml": export_cluster_yaml,
            "ansible_inventory": export_ansible_inventory,
        },
        "lists": {
            "rpms": export_packages_list,
            "hostname": export_hostname,
        },
        "nodes": export_nodes,
        "etcd": export_etcd,
        "cluster_version": export_kubernetes_version,
        "kubernetes": export_kubernetes,
    },
    "make_descriptor": make_descriptor,
    "pack": pack_data,
})


class BackupAction(Action):
    def __init__(self) -> None:
        super().__init__('backup')

    def run(self, res: DynamicResources) -> None:
        flow.run_tasks(res, tasks)


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for making backup of Kubernetes resources and nodes contents.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)

    context = flow.create_context(parser, cli_arguments, procedure='backup')
    context['execution_arguments']['disable_dump'] = False
    context['backup_descriptor'] = {
        'meta': {
            'time': {
                'started': datetime.datetime.now()
            }
        },
        'etcd': {},
        'nodes': {},
        'kubernetes': {
            'resources': {}
        }
    }

    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([BackupAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Simple TCP socket listener that can be run on both python 2 and 3,
# The listener accepts connections sequentially, and suppresses the received data.
# The script is for testing purpose only.
# The first argv parameter is the TCP port to listen. The second argv parameter is the ip protocol version.

import socket
import sys

if sys.argv[2] == '6' :
  s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
else:
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

try:
    s.bind(('', int(sys.argv[1])))
except socket.error as e:
    if "Address already in use" in str(e):
        sys.stdout.write("In use\n")
        sys.stdout.flush()
        exit(1)
    else:
        raise

sys.stdout.write("Listen\n")
sys.stdout.flush()
s.listen(1)

while True:
    client, _ = s.accept()
    while True:
        data = client.recv(1024)
        if not data:
            break

    client.close()



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from typing import Optional, List

from kubemarine import kubernetes, haproxy, keepalived
from kubemarine.core import flow, summary
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup
from kubemarine.core.resources import DynamicResources
from kubemarine.procedures import install


def _get_active_nodes(node_type: str, cluster: KubernetesCluster) -> Optional[NodeGroup]:
    all_nodes = None
    if cluster.nodes.get(node_type) is not None:
        all_nodes = cluster.nodes[node_type].get_nodes_for_removal()
    if all_nodes is None or all_nodes.is_empty():
        cluster.log.debug("Skipped - no %s to remove" % node_type)
        return None
    active_nodes = all_nodes.get_online_nodes(True)
    disabled_nodes = all_nodes.exclude_group(active_nodes)
    if active_nodes.is_empty():
        cluster.log.debug("Skipped - %s nodes are inactive: %s" % (node_type, ", ".join(disabled_nodes.nodes)))
        return None
    if not disabled_nodes.is_empty():
        cluster.log.debug("Partly Skipped - several %s nodes are inactive: %s"
                          % (node_type, ", ".join(disabled_nodes.nodes)))
    return active_nodes


def loadbalancer_remove_haproxy(cluster: KubernetesCluster) -> None:
    nodes = _get_active_nodes("balancer", cluster)
    if nodes is None:
        return
    nodes.call(haproxy.disable)


def loadbalancer_remove_keepalived(cluster: KubernetesCluster) -> None:
    nodes = _get_active_nodes("keepalived", cluster)
    if nodes is None:
        return
    nodes.call(keepalived.disable)


def remove_kubernetes_nodes(cluster: KubernetesCluster) -> None:
    group = cluster.make_group_from_roles(['control-plane', 'worker']).get_nodes_for_removal()

    if group.is_empty():
        cluster.log.debug("No kubernetes nodes to perform")
        return

    group.call(kubernetes.reset_installation_env)
    kubernetes.schedule_running_nodes_report(cluster)


def remove_node_finalize_inventory(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    if cluster.context.get('initial_procedure') != 'remove_node':
        return inventory_to_finalize

    nodes_for_removal = cluster.nodes['all'].get_nodes_for_removal()
    final_nodes = cluster.nodes['all'].get_final_nodes()

    # check if there are no more hosts where keepalived installed - remove according vrrp_ips
    for i, item in enumerate(inventory_to_finalize.get('vrrp_ips', [])):
        if 'hosts' in item:
            hosts = item['hosts']
        else:
            from kubemarine import keepalived
            hosts = keepalived.get_default_node_names(inventory_to_finalize)

        for host in hosts:
            host_name = host
            if isinstance(host_name, dict):
                host_name = host['name']
            if not final_nodes.has_node(host_name):
                hosts.remove(host)
        if not hosts:
            del inventory_to_finalize['vrrp_ips'][i]
        else:
            if inventory_to_finalize['vrrp_ips'][i].get('hosts', []):
                inventory_to_finalize['vrrp_ips'][i]['hosts'] = hosts

    # remove nodes from inventory if they in nodes for removal
    size = len(inventory_to_finalize['nodes'])
    for i in range(size):
        for j, node in enumerate(inventory_to_finalize['nodes']):
            if nodes_for_removal.has_node(node["name"]):
                del inventory_to_finalize['nodes'][j]
                break

    if inventory_to_finalize['services'].get('kubeadm', {}).get('apiServer', {}).get('certSANs'):
        for node in nodes_for_removal.get_ordered_members_configs_list():
            hostnames = [node['name'], node['internal_address']]
            if node.get('address') is not None:
                hostnames.append(node['address'])
            for name in hostnames:
                if name in inventory_to_finalize['services']['kubeadm']['apiServer']['certSANs']:
                    inventory_to_finalize['services']['kubeadm']['apiServer']['certSANs'].remove(name)

    return inventory_to_finalize


tasks = OrderedDict({
    "loadbalancer": {
        "remove": {
            "haproxy": loadbalancer_remove_haproxy,
            "keepalived": loadbalancer_remove_keepalived
        },
        "configure": {
            "haproxy": install.deploy_loadbalancer_haproxy_configure,
            "keepalived": install.deploy_loadbalancer_keepalived_configure
        },
    },
    "update": {
        "etc_hosts": install.system_prepare_dns_etc_hosts,
        "coredns": install.deploy_coredns
    },
    "remove_kubernetes_nodes": remove_kubernetes_nodes,
    "overview": install.overview,
})


cumulative_points = {
    summary.exec_delayed: [
        flow.END_OF_TASKS
    ]
}


class RemoveNodeAction(Action):
    def __init__(self) -> None:
        super().__init__('remove node', recreate_inventory=True)

    def run(self, res: DynamicResources) -> None:
        flow.run_tasks(res, tasks, cumulative_points=cumulative_points)
        res.make_final_inventory()


def main(cli_arguments: List[str] = None) -> None:

    cli_help = '''
    Script for removing node from Kubernetes cluster.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure='remove_node')

    flow.ActionsFlow([RemoveNodeAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
import jinja2

from kubemarine.core import defaults, log


def new(logger: log.EnhancedLogger, recursive_compile: bool = False, root: dict = None) -> jinja2.Environment:
    def _precompile(filter_: str, struct: str) -> str:
        if not isinstance(struct, str):
            raise ValueError(f"Filter {filter_!r} can be applied only on string")

        if not recursive_compile:
            return struct
        elif root is None:
            raise Exception(
                "If recursive compilation is enabled, "
                "'root' parameter should also be specified to provide compilation context.")

        struct = precompile(logger, struct, root)

        return struct


    env = jinja2.Environment()
    env.filters['toyaml'] = lambda data: yaml.dump(data, default_flow_style=False)
    env.filters['isipv4'] = lambda ip: ":" not in _precompile('isipv4', ip)
    env.filters['minorversion'] = lambda version: ".".join(_precompile('minorversion', version).split('.')[0:2])
    env.filters['majorversion'] = lambda version: _precompile('majorversion', version).split('.')[0]
    # we need these filters because rendered cluster.yaml can contain variables like 
    # enable: 'true'
    env.filters['is_true'] = lambda v: v in ['true', 'True', 'TRUE', True]
    env.filters['is_false'] = lambda v: v in ['false', 'False', 'FALSE', False]
    return env


def precompile(logger: log.EnhancedLogger, struct: str, root: dict) -> str:
    # maybe we have non compiled string like templates/plugins/calico-{{ globals.compatibility_map }} ?
    if '{{' in struct or '{%' in struct:
        struct = defaults.compile_string(logger, struct, root)

    return struct



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import shlex
import sys
import time
from abc import abstractmethod, ABC
from copy import deepcopy
from types import FunctionType
from typing import Type, Optional, List, Union, Sequence, Tuple, cast, Callable, Dict, Any

from kubemarine.core import utils, cluster as c, action, resources as res, errors, summary, log

DEFAULT_CLUSTER_OBJ: Optional[Type[c.KubernetesCluster]] = None
TASK_DESCRIPTION_TEMPLATE = """
tasks list:
    %s
"""

END_OF_TASKS = object()


class FlowResult:
    def __init__(self, context: dict, logger: log.EnhancedLogger):
        self.context = context
        self.logger = logger


class Flow(ABC):
    def run_flow(self, context: Union[dict, res.DynamicResources], print_summary: bool = True) -> FlowResult:
        time_start = time.time()

        if isinstance(context, res.DynamicResources):
            resources = context
        else:
            resources = res.DynamicResources(context)

        context = resources.context
        args: dict = context['execution_arguments']

        try:
            if not args['disable_dump']:
                utils.prepare_dump_directory(args['dump_location'],
                                             reset_directory=not args['disable_dump_cleanup'])
            resources.logger()
            self._run(resources)
        except Exception as exc:
            logger = resources.logger_if_initialized()
            if isinstance(exc, errors.FailException):
                utils.do_fail(exc.message, exc.reason, exc.hint, logger=logger)
            else:
                utils.do_fail(f"'{context['initial_procedure'] or 'undefined'}' procedure failed.", exc,
                              logger=logger)

        time_end = time.time()
        logger = resources.logger()

        if print_summary:
            summary.schedule_report(resources.working_context, summary.SummaryItem.EXECUTION_TIME,
                                    utils.get_elapsed_string(time_start, time_end))
            summary.print_summary(resources.working_context, logger)
            logger.info("SUCCESSFULLY FINISHED")

        return FlowResult(resources.working_context, logger)

    @abstractmethod
    def _run(self, resources: res.DynamicResources) -> None:
        pass


class ActionsFlow(Flow):
    def __init__(self, actions: List[action.Action]):
        self._actions = actions

    def _run(self, resources: res.DynamicResources) -> None:
        run_actions(resources, self._actions)


def run_actions(resources: res.DynamicResources, actions: Sequence[action.Action]) -> None:
    """
    Runs actions one by one, recreates inventory when necessary,
    managing such resources as cluster object and raw inventory.

    For each initialized cluster object, preserves inventory if any action is succeeded.
    """

    context = resources.context
    logger = resources.logger()

    successfully_performed: List[str] = []
    last_cluster = None
    for act in actions:
        act.prepare_context(context)

        if not successfully_performed:
            # first action in group
            if resources.inventory_filepath:
                with utils.open_external(resources.inventory_filepath, "r") as stream:
                    utils.dump_file(context, stream, "cluster_initial.yaml")

            if resources.procedure_inventory_filepath:
                with utils.open_external(resources.procedure_inventory_filepath, "r") as stream:
                    utils.dump_file(context, stream, "procedure.yaml")
        try:
            logger.info(f"Running action '{act.identifier}'")
            act.run(resources)
            act.reset_context(context)
            successfully_performed.append(act.identifier)
        except Exception:
            if successfully_performed:
                _post_process_actions_group(last_cluster, context, successfully_performed, failed=True)

            raise

        last_cluster = resources.cluster_if_initialized()

        if act.recreate_inventory:
            if resources.inventory_filepath:
                with utils.open_external(resources.inventory_filepath, "r") as stream:
                    # write original file data to backup file with timestamp
                    timestamp = utils.get_current_timestamp_formatted()
                    inventory_file_basename = os.path.basename(resources.inventory_filepath)
                    utils.dump_file(context, stream, "%s_%s" % (inventory_file_basename, str(timestamp)))

            resources.recreate_inventory()
            _post_process_actions_group(last_cluster, context, successfully_performed)
            successfully_performed = []
            last_cluster = None

    if successfully_performed:
        _post_process_actions_group(last_cluster, context, successfully_performed)


def _post_process_actions_group(last_cluster: Optional[c.KubernetesCluster], context: dict,
                                successfully_performed: list, failed: bool = False) -> None:
    if last_cluster is None:
        return
    try:
        last_cluster.dump_finalized_inventory()
    finally:
        if context['preserve_inventory']:
            last_cluster.context['successfully_performed'] = successfully_performed
            last_cluster.context['status'] = 'failed' if failed else 'successful'
            last_cluster.preserve_inventory()


def run_tasks(resources: res.DynamicResources, tasks: dict, cumulative_points: dict = None,
              tasks_filter: List[str] = None) -> None:
    """
    Filters and runs tasks.
    """

    if cumulative_points is None:
        cumulative_points = {}

    args: dict = resources.context['execution_arguments']

    tasks_filter = tasks_filter if tasks_filter is not None \
        else [] if not args.get('tasks') else args['tasks'].split(",")
    excluded_tasks = [] if not args.get('exclude') else args['exclude'].split(",")

    print("Excluded tasks:")
    filtered_tasks, final_list = filter_flow(tasks, tasks_filter, excluded_tasks)
    if filtered_tasks == tasks:
        print("\tNo excluded tasks")

    cluster = resources.cluster()

    if args.get('without_act', False):
        resources.context['preserve_inventory'] = False
        cluster.log.debug('\nFurther acting manually disabled')
        return

    init_tasks_flow(cluster)
    run_tasks_recursive(tasks, final_list, cluster, cumulative_points, [])
    proceed_cumulative_point(cluster, cumulative_points, END_OF_TASKS,
                             force=args.get('force_cumulative_points', False))


def create_empty_context(args: dict = None, procedure: str = None) -> dict:
    if args is None:
        args = {}
    return {
        "execution_arguments": deepcopy(args),
        "nodes": {},
        'initial_procedure': procedure,
        'preserve_inventory': True,
        'runtime_vars': {}
    }


def get_task_list(tasks: dict, _task_path: str = '') -> List[str]:
    result = []
    for task_name, task in tasks.items():
        __task_path = _task_path + "." + task_name if _task_path != '' else task_name
        result.extend(get_task_list(task, __task_path) if not callable(task) else [__task_path])
    return result


def create_context(parser: argparse.ArgumentParser, cli_arguments: Optional[list], procedure: str) -> dict:
    args_list = sys.argv[1:]
    if cli_arguments is not None:
        args_list = cli_arguments

    parser.prog = procedure
    args = vars(parse_args(parser, args_list))

    if args.get('exclude_cumulative_points_methods', '').strip() != '':
        args['exclude_cumulative_points_methods'] = args['exclude_cumulative_points_methods'].strip().split(",")
        # print('The following cumulative points methods are marked for exclusion: [ %s ]' %
        #               ', '.join(args['exclude_cumulative_points_methods']))
    else:
        args['exclude_cumulative_points_methods'] = []

    context = create_empty_context(args=args, procedure=procedure)
    context["initial_cli_arguments"] = ' '.join(map(shlex.quote, args_list))

    return context


def filter_flow(tasks: dict, tasks_filter: List[str], excluded_tasks: List[str]) -> Tuple[dict, List[str]]:
    # Remove any whitespaces from filters, and split by '.'
    tasks_path_filter = [tasks.split(".") for tasks in list(map(str.strip, tasks_filter))]
    excluded_path_tasks = [tasks.split(".") for tasks in list(map(str.strip, excluded_tasks))]

    return _filter_flow_internal(tasks, tasks_path_filter, excluded_path_tasks, [])


def _filter_flow_internal(tasks: dict, tasks_filter: List[List[str]], excluded_tasks: List[List[str]],
                          _task_path: List[str]) -> Tuple[dict, List[str]]:
    filtered = {}
    final_list = []

    for task_name, task in tasks.items():
        __task_path = _task_path + [task_name]
        __task_name = ".".join(__task_path)

        allowed = True
        # if task_filter is not empty - smb specified filter argument
        if tasks_filter:
            allowed = False
            # Check if the iterable subpath is in allowed paths. For example we have to check if
            # system_prepare.cri in allowed path system_prepare.cri.docker
            for task_path in tasks_filter:
                # one of task_path, __task_path is a sublist of another
                # check if current '__task_path' is a sublist only if 'task' is not a final task.
                if (task_path[:len(__task_path)] == __task_path and not callable(task)) \
                        or __task_path[:len(task_path)] == task_path:
                    allowed = True

        if allowed and __task_path not in excluded_tasks:
            if callable(task):
                filtered[task_name] = task
                final_list.append(__task_name)
            else:
                filtered_flow, _final_list = _filter_flow_internal(task, tasks_filter, excluded_tasks, __task_path)
                # there is something to execute in subtree
                if filtered_flow:
                    filtered[task_name] = filtered_flow
                    final_list += _final_list
        else:
            print("\t%s" % __task_name)

    return filtered, final_list


def run_tasks_recursive(tasks: dict, final_task_names: List[str], cluster: c.KubernetesCluster,
                        cumulative_points: dict, _task_path: List[str]) -> None:
    for task_name, task in tasks.items():
        __task_path = _task_path + [task_name]
        __task_name = ".".join(__task_path)
        run = __task_name in final_task_names

        args = cluster.context['execution_arguments']
        # --force-cumulative-points forcibly run the point only if the related task is going to be executed
        force_cumulative_point = run and args.get('force_cumulative_points', False)
        proceed_cumulative_point(cluster, cumulative_points, __task_name, force=force_cumulative_point)

        if callable(task):
            if not run:
                continue
            cluster.log.info("*** TASK %s ***" % __task_name)
            try:
                task(cluster)
                add_task_to_proceeded_list(cluster, __task_name)
            except Exception as exc:
                raise errors.FailException(
                    "TASK FAILED %s" % __task_name, exc,
                    hint=cluster.globals['error_handling']['failure_message'] % (sys.argv[0], __task_name)
                )
        else:
            run_tasks_recursive(task, final_task_names, cluster, cumulative_points, __task_path)


def new_common_parser(cli_help: str) -> argparse.ArgumentParser:

    parser = argparse.ArgumentParser(description=cli_help,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-c', '--config',
                        default='cluster.yaml',
                        help='define main cluster configuration file')

    parser.add_argument('--ansible-inventory-location',
                        default='./ansible-inventory.ini',
                        help='auto-generated ansible-compatible inventory file location')

    parser.add_argument('--dump-location',
                        default='./',
                        help='dump directory for intermediate files')

    parser.add_argument('--disable-dump',
                        action='store_true',
                        help='prevent dump directory creation')

    parser.add_argument('--disable-dump-cleanup',
                        action='store_true',
                        help='prevent dump directory cleaning on process launch')

    parser.add_argument('--log',
                        action='append',
                        nargs='*',
                        help='Logging options, can be specified multiple times')

    parser.add_argument('-w', '--workdir',
                        default='',
                        help='Custom path of the workdir')

    return parser


def new_tasks_flow_parser(cli_help: str, tasks: dict = None) -> argparse.ArgumentParser:
    parser = new_common_parser(cli_help)

    parser.add_argument('--without-act',
                        action='store_true',
                        help='prevent tasks to be executed')

    parser.add_argument('--tasks',
                        default='',
                        help='define comma-separated tasks to be executed')

    parser.add_argument('--exclude',
                        default='',
                        help='exclude comma-separated tasks from execution')

    parser.add_argument('--disable-cumulative-points',
                        action='store_true',
                        help='disable cumulative points execution (use only when you understand what you are doing!)')

    parser.add_argument('--force-cumulative-points',
                        action='store_true',
                        help='force cumulative points execution - they will be executed regardless of whether it was '
                             'scheduled or not (use only when you understand what you are doing!)')

    parser.add_argument('--exclude-cumulative-points-methods',
                        default='',
                        help='comma-separated cumulative points methods names to be excluded from execution')

    # Add tasks list to help section
    if tasks is not None:
        parser.epilog = TASK_DESCRIPTION_TEMPLATE % ('\n    '.join(get_task_list(tasks)))

    return parser


def new_procedure_parser(cli_help: str, optional_config: bool = False, tasks: dict = None) -> argparse.ArgumentParser:
    parser = new_tasks_flow_parser(cli_help, tasks)

    help_msg = 'config file for the procedure'
    if optional_config:
        parser.add_argument('procedure_config', metavar='procedure_config', type=str, help=help_msg, nargs='?')
    else:
        parser.add_argument('procedure_config', metavar='procedure_config', type=str, help=help_msg)

    return parser


def parse_args(parser: argparse.ArgumentParser, arguments: list) -> argparse.Namespace:
    args = parser.parse_args(arguments)

    if args.workdir != '':
        os.chdir(args.workdir)

    return args


def schedule_cumulative_point(cluster: c.KubernetesCluster, point_method: Callable) -> None:
    _check_within_flow(cluster)

    func = cast(FunctionType, point_method)
    point_fullname = func.__module__ + '.' + func.__qualname__

    if cluster.context['execution_arguments'].get('disable_cumulative_points', False):
        cluster.log.verbose('Method %s not scheduled - cumulative points disabled' % point_fullname)
        return

    if point_fullname in cluster.context['execution_arguments']['exclude_cumulative_points_methods']:
        cluster.log.verbose('Method %s not scheduled - it set to be excluded' % point_fullname)
        return

    scheduled_points = cluster.context.get('scheduled_cumulative_points', [])

    if point_method not in scheduled_points:
        scheduled_points.append(point_method)
        cluster.context['scheduled_cumulative_points'] = scheduled_points
        cluster.log.verbose('Method %s scheduled' % point_fullname)
    else:
        cluster.log.verbose('Method %s already scheduled' % point_fullname)


def proceed_cumulative_point(cluster: c.KubernetesCluster, points_list: dict,
                             point_task_name: Union[str, object], force: bool = False) -> Dict[str, Any]:
    _check_within_flow(cluster)

    if cluster.context['execution_arguments'].get('disable_cumulative_points', False):
        return {}

    scheduled_methods = cluster.context.get('scheduled_cumulative_points', [])

    results = {}
    for point_method, points_tasks_names in points_list.items():
        if point_task_name in points_tasks_names:

            point_method_fullname = point_method.__module__ + '.' + point_method.__qualname__
            if force:
                cluster.log.verbose('Method %s will be forcibly executed' % point_method_fullname)
            elif point_method not in scheduled_methods:
                cluster.log.verbose('Method %s not scheduled - cumulative point call skipped' % point_method_fullname)
                continue

            cluster.log.info("*** CUMULATIVE POINT %s ***" % point_method_fullname)

            call_result = point_method(cluster)
            if point_method in scheduled_methods:
                scheduled_methods.remove(point_method)
            results[point_method_fullname] = call_result

    return results


def init_tasks_flow(cluster: c.KubernetesCluster) -> None:
    if 'proceeded_tasks' not in cluster.context:
        cluster.context['proceeded_tasks'] = []


def add_task_to_proceeded_list(cluster: c.KubernetesCluster, task_path: str) -> None:
    if not is_task_completed(cluster, task_path):
        cluster.context['proceeded_tasks'].append(task_path)
        utils.dump_file(cluster, "\n".join(cluster.context['proceeded_tasks'])+"\n", 'finished_tasks')


def is_task_completed(cluster: c.KubernetesCluster, task_path: str) -> bool:
    _check_within_flow(cluster)
    return task_path in cluster.context['proceeded_tasks']


def _check_within_flow(cluster: c.KubernetesCluster, check: bool = True) -> None:
    if check != ('proceeded_tasks' in cluster.context):
        raise NotImplementedError(f"The method is called {'not ' if check else ''}within tasks flow execution")



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Optional

from kubemarine.core import summary, utils, log
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest


def enrich_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if not inventory["plugins"]["kubernetes-dashboard"]["install"]:
        return inventory

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_dashboard = cluster.raw_inventory.get("plugins", {}).get("kubernetes-dashboard", {}).get("dashboard", {})
    if "resources" in raw_dashboard:
        inventory["plugins"]["kubernetes-dashboard"]["dashboard"]["resources"] = raw_dashboard["resources"]
    raw_metrics_scrapper = cluster.raw_inventory.get("plugins", {}).get("kubernetes-dashboard", {}).get("metrics-scraper", {})
    if "resources" in raw_metrics_scrapper:
        inventory["plugins"]["kubernetes-dashboard"]["metrics-scraper"]["resources"] = raw_metrics_scrapper["resources"]

    return inventory


def schedule_summary_report(cluster: KubernetesCluster) -> None:
    plugin_item = cluster.inventory['plugins']['kubernetes-dashboard']
    hostname = plugin_item['hostname']
    # Currently we declare that Dashboard UI is available only via HTTPS
    summary.schedule_report(cluster.context, summary.SummaryItem.DASHBOARD_URL, f'https://{hostname}')


class DashboardManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None) -> None:
        super().__init__(logger, inventory, 'kubernetes-dashboard', original_yaml_path, destination_name)

    def get_known_objects(self) -> List[str]:
        return [
            "Namespace_kubernetes-dashboard",
            "ServiceAccount_kubernetes-dashboard",
            "Service_kubernetes-dashboard",
            "Secret_kubernetes-dashboard-certs",
            "Secret_kubernetes-dashboard-csrf",
            "Secret_kubernetes-dashboard-key-holder",
            "ConfigMap_kubernetes-dashboard-settings",
            "Role_kubernetes-dashboard",
            "ClusterRole_kubernetes-dashboard",
            "RoleBinding_kubernetes-dashboard",
            "ClusterRoleBinding_kubernetes-dashboard",
            "Deployment_kubernetes-dashboard",
            "Service_dashboard-metrics-scraper",
            "Deployment_dashboard-metrics-scraper",
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.enrich_namespace_kubernetes_dashboard,
            self.enrich_deployment_kubernetes_dashboard,
            self.enrich_deployment_dashboard_metrics_scraper,
        ]

    def enrich_namespace_kubernetes_dashboard(self, manifest: Manifest) -> None:
        key = "Namespace_kubernetes-dashboard"
        rbac = self.inventory['rbac']
        if rbac['admission'] == 'pss' and rbac['pss']['pod-security'] == 'enabled' \
                and rbac['pss']['defaults']['enforce'] == 'restricted':
            self.assign_default_pss_labels(manifest, key, 'baseline')

    def enrich_deployment_kubernetes_dashboard(self, manifest: Manifest) -> None:
        key = "Deployment_kubernetes-dashboard"
        self.enrich_image_for_container(manifest, key,
            plugin_service='dashboard', container_name='kubernetes-dashboard', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='kubernetes-dashboard', plugin_service="dashboard")
        self.enrich_node_selector(manifest, key, plugin_service='dashboard')
        self.enrich_tolerations(manifest, key, plugin_service='dashboard', override=True)

    def enrich_deployment_dashboard_metrics_scraper(self, manifest: Manifest) -> None:
        key = "Deployment_dashboard-metrics-scraper"
        self.enrich_image_for_container(manifest, key,
            plugin_service='metrics-scraper', container_name='dashboard-metrics-scraper', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='dashboard-metrics-scraper', plugin_service="metrics-scraper")
        self.enrich_node_selector(manifest, key, plugin_service='metrics-scraper')
        self.enrich_tolerations(manifest, key, plugin_service='metrics-scraper', override=True)


class V2_5_X_DashboardManifestProcessor(DashboardManifestProcessor):
    def enrich_deployment_dashboard_metrics_scraper(self, manifest: Manifest) -> None:
        key = "Deployment_dashboard-metrics-scraper"
        source_yaml = manifest.get_obj(key, patch=True)
        template_spec: dict = source_yaml['spec']['template']['spec']
        del template_spec['securityContext']
        self.log.verbose(f"The 'securityContext' property has been removed from 'spec.template.spec' in the {key}")
        super().enrich_deployment_dashboard_metrics_scraper(manifest)


def get_dashboard_manifest_processor(logger: log.VerboseLogger, inventory: dict,
                                     yaml_path: Optional[str] = None, destination: Optional[str] = None) -> Processor:
    version: str = inventory['plugins']['kubernetes-dashboard']['version']
    kwargs = {'original_yaml_path': yaml_path, 'destination_name': destination}
    if utils.minor_version(version) == 'v2.5':
        return V2_5_X_DashboardManifestProcessor(logger, inventory, **kwargs)

    return DashboardManifestProcessor(logger, inventory, **kwargs)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ipaddress
from typing import Optional, List, Dict

import os

from kubemarine.core import utils, log
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest


def enrich_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if not inventory["plugins"]["calico"]["install"]:
        return inventory

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_calico_node = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("node", {})
    if "resources" in raw_calico_node:
        inventory["plugins"]["calico"]["node"]["resources"] = raw_calico_node["resources"]
    raw_typha = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("typha", {})
    if "resources" in raw_typha:
        inventory["plugins"]["calico"]["typha"]["resources"] = raw_typha["resources"]
    raw_calico_controller = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("kube-controllers", {})
    if "resources" in raw_calico_controller:
        inventory["plugins"]["calico"]["kube-controllers"]["resources"] = raw_calico_controller["resources"]

    return inventory


# DEPRECATED
def apply_calico_yaml(cluster: KubernetesCluster, calico_original_yaml: str, calico_yaml: str) -> None:
    """
    The method implements full processing for Calico plugin
    :param calico_original_yaml: path to original Calico manifest
    :param calico_yaml: file name of the resulting Calico manifest
    :param cluster: Cluster object
    """

    calico_yaml = os.path.basename(calico_yaml)
    processor = CalicoManifestProcessor(cluster.log, cluster.inventory,
                                        original_yaml_path=calico_original_yaml,
                                        destination_name=calico_yaml)
    manifest = processor.enrich()
    processor.apply(cluster, manifest)


class CalicoManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None):
        super().__init__(logger, inventory, 'calico', original_yaml_path, destination_name)

    def exclude_typha_objects_if_disabled(self, manifest: Manifest) -> None:
        # enrich 'calico-typha' objects only if it's enabled in 'cluster.yaml'
        # in other case those objects must be excluded
        str_value = utils.true_or_false(self.inventory['plugins']['calico']['typha']['enabled'])
        if str_value == 'false':
            for key in ("Deployment_calico-typha", "Service_calico-typha", "PodDisruptionBudget_calico-typha"):
                self.exclude(manifest, key)
        elif str_value == 'true':
            return
        else:
            raise Exception(f"plugins.calico.typha.enabled must be set in 'True' or 'False' "
                            f"as string or boolean value")

    def enrich_configmap_calico_config(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico ConfigMap
        :param manifest: Container to operate with manifest objects
        """

        key = "ConfigMap_calico-config"
        source_yaml = manifest.get_obj(key, patch=True)
        val = self.inventory['plugins']['calico']['mtu']
        source_yaml['data']['veth_mtu'] = str(val)
        self.log.verbose(f"The {key} has been patched in 'data.veth_mtu' with '{val}'")
        str_value = utils.true_or_false(self.inventory['plugins']['calico']['typha']['enabled'])
        if str_value == "true":
            val = "calico-typha"
        elif str_value == "false":
            val = "none"
        source_yaml['data']['typha_service_name'] = val
        self.log.verbose(f"The {key} has been patched in 'data.typha_service_name' with '{val}'")
        string_part = source_yaml['data']['cni_network_config']
        ip = self.inventory['services']['kubeadm']['networking']['podSubnet'].split('/')[0]
        if type(ipaddress.ip_address(ip)) is ipaddress.IPv4Address:
            val = self.inventory['plugins']['calico']['cni']['ipam']['ipv4']
        else:
            val = self.inventory['plugins']['calico']['cni']['ipam']['ipv6']
        new_string_part = string_part.replace('"type": "calico-ipam"', str(val)[:-1][1:].replace("'", "\""))
        source_yaml['data']['cni_network_config'] = new_string_part
        log_str = new_string_part.replace("\n", "")
        self.log.verbose(f"The {key} has been patched in 'data.cni_network_config' with '{log_str}'")

    def enrich_deployment_calico_kube_controllers(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico controller Deployment
        :param manifest: Container to operate with manifest objects
        """

        key = "Deployment_calico-kube-controllers"
        self.enrich_node_selector(manifest, key, plugin_service='kube-controllers')
        self.enrich_resources_for_container(manifest, key,
            plugin_service='kube-controllers', container_name='calico-kube-controllers')
        self.enrich_image_for_container(manifest, key,
            plugin_service='kube-controllers', container_name='calico-kube-controllers', is_init_container=False)

    def enrich_daemonset_calico_node(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico node DaemonSet
        :param manifest: Container to operate with manifest objects
        """

        key = "DaemonSet_calico-node"
        for container_name in ['upgrade-ipam', 'install-cni']:
            self.enrich_image_for_container(manifest, key,
                plugin_service='cni', container_name=container_name, is_init_container=True)

        self.enrich_image_for_container(manifest, key,
            plugin_service='node', container_name='mount-bpffs', is_init_container=True, allow_absent=True)
        self.enrich_image_for_container(manifest, key,
            plugin_service='flexvol', container_name='flexvol-driver', is_init_container=True, allow_absent=True)

        self.enrich_image_for_container(manifest, key,
            plugin_service='node', container_name='calico-node', is_init_container=False)

        container_pos, container = self.find_container_for_patch(manifest, key,
            container_name='calico-node', is_init_container=False)

        self.enrich_daemonset_calico_node_container_env(container_pos, container)
        self.enrich_resources_for_container(manifest, key,
            plugin_service='node', container_name='calico-node')

    def enrich_daemonset_calico_node_container_env(self, container_pos: int, container: dict) -> None:
        """
        The method implements the enrichment procedure for 'calico-node' container in Calico node DaemonSet.
        The method attempts to preserve initial formatting.

        :param container_pos: container position in spec
        :param container: object describing a container
        """
        key = "DaemonSet_calico-node"
        env_delete: List[str] = []
        ip = self.inventory['services']['kubeadm']['networking']['podSubnet'].split('/')[0]
        if type(ipaddress.ip_address(ip)) is ipaddress.IPv4Address:
            env_delete.extend([
                'CALICO_IPV6POOL_CIDR', 'IP6', 'IP6_AUTODETECTION_METHOD',
                'CALICO_IPV6POOL_IPIP', 'CALICO_IPV6POOL_VXLAN'
            ])
        if utils.true_or_false(self.inventory['plugins']['calico']['typha']['enabled']) == "false":
            env_delete.append('FELIX_TYPHAK8SSERVICENAME')

        for name in env_delete:
            for i, e in enumerate(container['env']):
                if e['name'] == name:
                    del container['env'][i]
                    self.log.verbose(f"The {name!r} env variable has been removed from "
                                    f"'spec.template.spec.containers.[{container_pos}].env' in the {key}")
                    break

        env_update: Dict[str, dict] = {}
        for name, value in self.inventory['plugins']['calico']['env'].items():
            if name in env_delete:
                continue
            if type(value) is str:
                env_update[name] = {'value': value}
            elif type(value) is dict:
                env_update[name] = {'valueFrom': value}
            self.log.verbose(f"The {key} has been patched in "
                            f"'spec.template.spec.containers.[{container_pos}].env.{name}' with '{value}'")

        for env in container['env']:
            name = env['name']
            if name not in env_update:
                continue

            value = env_update.pop(name)
            keys = list(env.keys())
            for key in keys:
                if key != 'name' and key not in value:
                    del env[key]

            env.update(value)

        for name, env in env_update.items():
            new_env = {'name' : name}
            new_env.update(env)
            container['env'].append(new_env)

    def enrich_deployment_calico_typha(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Typha Deployment
        :param manifest: Container to operate with manifest objects
        """

        key = "Deployment_calico-typha"
        if not manifest.has_obj(key):
            return None
        source_yaml = manifest.get_obj(key, patch=True)

        default_tolerations = [{'key': 'node.kubernetes.io/network-unavailable', 'effect': 'NoSchedule'},
                               {'key': 'node.kubernetes.io/network-unavailable', 'effect': 'NoExecute'}]

        val = self.inventory['plugins']['calico']['typha']['replicas']
        source_yaml['spec']['replicas'] = int(val)
        self.log.verbose(f"The {key} has been patched in 'spec.replicas' with '{val}'")

        self.enrich_node_selector(manifest, key, plugin_service='typha')
        self.enrich_tolerations(manifest, key, plugin_service='typha', extra_tolerations=default_tolerations)
        self.enrich_image_for_container(manifest, key,
            plugin_service='typha', container_name='calico-typha', is_init_container=False)
        self.enrich_resources_for_container(manifest, key,
            plugin_service='typha', container_name='calico-typha')


    def enrich_clusterrole_calico_kube_controllers(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico controller ClusterRole
        :param manifest: Container to operate with manifest objects
        """

        key = "ClusterRole_calico-kube-controllers"
        if self.inventory['rbac']['admission'] == "psp" and \
                self.inventory['rbac']['psp']['pod-security'] == "enabled":
            source_yaml = manifest.get_obj(key, patch=True)
            api_list = source_yaml['rules']
            api_list.append(psp_calico_kube_controllers)
            self.log.verbose(f"The {key} has been patched in 'rules' with '{psp_calico_kube_controllers}'")

    def enrich_clusterrole_calico_node(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico node ClusterRole
        :param manifest: Container to operate with manifest objects
        """

        key = "ClusterRole_calico-node"
        if self.inventory['rbac']['admission'] == "psp" and \
                self.inventory['rbac']['psp']['pod-security'] == "enabled":
            source_yaml = manifest.get_obj(key, patch=True)
            api_list = source_yaml['rules']
            api_list.append(psp_calico_node)
            self.log.verbose(f"The {key} has been patched in 'rules' with '{psp_calico_node}'")

    def enrich_crd_felix_configuration(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico CRD Felixconfigurations
        :param manifest: Container to operate with manifest objects
        """

        key = "CustomResourceDefinition_felixconfigurations.crd.projectcalico.org"
        source_yaml = manifest.get_obj(key, patch=True)

        api_list = \
        source_yaml['spec']['versions'][0]['schema']['openAPIV3Schema']['properties']['spec']['properties'][
            'prometheusMetricsEnabled']
        api_list["default"] = True
        source_yaml['spec']['versions'][0]['schema']['openAPIV3Schema']['properties']['spec']['properties'][
            'prometheusMetricsEnabled'] = api_list

        sz = len(manifest.all_obj_keys())
        import ruamel.yaml
        self.include(manifest, sz, ruamel.yaml.safe_load(utils.read_internal('templates/plugins/calico-kube-controllers-metrics.yaml')))
        self.include(manifest, sz, ruamel.yaml.safe_load(utils.read_internal('templates/plugins/calico-metrics.yaml')))

    def get_known_objects(self) -> List[str]:
        return [
            "ConfigMap_calico-config",
            "CustomResourceDefinition_bgpconfigurations.crd.projectcalico.org",
            "CustomResourceDefinition_bgppeers.crd.projectcalico.org",
            "CustomResourceDefinition_blockaffinities.crd.projectcalico.org",
            "CustomResourceDefinition_caliconodestatuses.crd.projectcalico.org",
            "CustomResourceDefinition_clusterinformations.crd.projectcalico.org",
            "CustomResourceDefinition_felixconfigurations.crd.projectcalico.org",
            "CustomResourceDefinition_globalnetworkpolicies.crd.projectcalico.org",
            "CustomResourceDefinition_globalnetworksets.crd.projectcalico.org",
            "CustomResourceDefinition_hostendpoints.crd.projectcalico.org",
            "CustomResourceDefinition_ipamblocks.crd.projectcalico.org",
            "CustomResourceDefinition_ipamconfigs.crd.projectcalico.org",
            "CustomResourceDefinition_ipamhandles.crd.projectcalico.org",
            "CustomResourceDefinition_ippools.crd.projectcalico.org",
            "CustomResourceDefinition_ipreservations.crd.projectcalico.org",
            "CustomResourceDefinition_kubecontrollersconfigurations.crd.projectcalico.org",
            "CustomResourceDefinition_networkpolicies.crd.projectcalico.org",
            "CustomResourceDefinition_networksets.crd.projectcalico.org",
            "ClusterRole_calico-kube-controllers",
            "ClusterRoleBinding_calico-kube-controllers",
            "ClusterRole_calico-node",
            "ClusterRoleBinding_calico-node",
            "DaemonSet_calico-node",
            "ServiceAccount_calico-node",
            "Deployment_calico-kube-controllers",
            "ServiceAccount_calico-kube-controllers",
            "PodDisruptionBudget_calico-kube-controllers",
            "Deployment_calico-typha",
            "Service_calico-typha",
            "PodDisruptionBudget_calico-typha"
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.exclude_typha_objects_if_disabled,
            self.enrich_configmap_calico_config,
            self.enrich_deployment_calico_kube_controllers,
            self.enrich_daemonset_calico_node,
            self.enrich_deployment_calico_typha,
            self.enrich_clusterrole_calico_kube_controllers,
            self.enrich_clusterrole_calico_node,
            self.enrich_crd_felix_configuration,
        ]


psp_calico_kube_controllers = {
        "apiGroups": ["policy"],
        "resources": ["podsecuritypolicies"],
        "verbs":     ["use"],
        "resourceNames": ["oob-anyuid-psp"]
}

psp_calico_node = {
        "apiGroups": ["policy"],
        "resources": ["podsecuritypolicies"],
        "verbs":     ["use"],
        "resourceNames": ["oob-privileged-psp"]
}
